@@ -4,6 +4,7 @@ import json
 import matplotlib.pyplot as plt
 import numpy as np
 import os
+import shapely.geometry
 from shapely import ops, wkt
 from shapely.geometry import Polygon, LineString, Point
 
@@ -11,6 +12,8 @@ from matplotlib import patches
 from pyproj import CRS, Transformer
 from pyproj.exceptions import CRSError
 
+from typing import Any, Dict, List, Optional, Tuple, Union
+
 from .. import cv, helpers
 
 
@@ -27,20 +30,22 @@ class CameraConfig:
 
     def __init__(
             self,
-            height,
-            width,
-            crs=None,
-            window_size=10,
-            resolution=0.05,
-            bbox=None,
-            camera_matrix=None,
-            dist_coeffs=None,
-            lens_position=None,
-            corners=None,
-            gcps=None,
-            lens_pars=None,
-            calibration_video=None
-
+            height: int,
+            width: int,
+            crs: Optional[Any] = None,
+            window_size: int = 10,
+            resolution: float = 0.05,
+            bbox: Optional[Union[shapely.geometry.Polygon, str]] = None,
+            camera_matrix: Optional[List[List[float]]] = None,
+            dist_coeffs: Optional[List[List[float]]] = None,
+            lens_position: Optional[List[float]] = None,
+            corners: Optional[List[List[float]]] = None,
+            gcps: Optional[Dict[str, Union[List, float]]] = None,
+            lens_pars: Optional[Dict[str, float]] = None,
+            calibration_video: Optional[str] = None,
+            is_nadir: Optional[bool] = False,
+            stabilize: Optional[List[List]] = None,
+            rotation: Optional[int] = None,
     ):
         """
 
@@ -64,52 +69,70 @@ class CameraConfig:
         corners : list of lists of floats (2)
             [x, y] coordinates defining corners of area of interest in camera cols/rows, bbox will be computed from this
         gcps : dict
-            Can contain "src": list of 4 lists, with column, row locations in objective of control points,
-            "dst": list of 4 lists, with x, y locations (local or global coordinate reference system) of control points,
+            Can contain "src": list of lists, with column, row locations in objective of control points,
+            "dst": list of lists, with x, y or x, y, z locations (local or global coordinate reference system) of
+            control points,
             "h_ref": float, measured water level [m] in local reference system (e.g. from staff gauge or pressure gauge)
             during gcp survey,
             "z_0": float, water level [m] in global reference system (e.g. from used GPS system CRS). This must be in
             the same vertical reference as the measured bathymetry and other survey points,
             "crs": int, str or CRS object, CRS in which "dst" points are measured. If None, a local coordinate system is
             assumed (e.g. from spirit level).
-        lens_pars : dict, optional
+        lens_pars (deprecated) : dict, optional
             Lens parameters, containing: "k1": float, barrel lens distortion parameter (default: 0.),
             "c": float, optical center (default: 2.),
-            "f": float, focal length (default: 1.)
+            "focal_length": float, focal length (default: width of image frame)
         calibration_video : str, optional
             local path to video file containing a checkerboard pattern. Must be 9x6 if called directly, otherwise use
             ``.calibrate_camera`` explicitly and provide ``chessboard_size`` explicitly. When used, an automated camera
             calibration on the video file will be attempted.
+        is_nadir : bool, optional
+            If set, the video is assumed to be taken at sub-drone position and only two control points are needed
+            for camera configuration
+        stabilize : list (of lists), optional
+            contains [x, y] pixels defining a polygon enclosing moving (water) areas. Areas outside of the polygon
+            are used for stabilization of the video, if this polygon is defined.
+        rotation : int [90, 180, 270]
+            enforces a rotation of the video of 90, 180 or 2780 degrees clock-wise.
         """
         assert(isinstance(height, int)), 'height must be provided as type "int"'
         assert(isinstance(width, int)), 'width must be provided as type "int"'
         assert (isinstance(window_size, int)), 'window_size must be of type "int"'
         self.height = height
         self.width = width
+        self.is_nadir = is_nadir
         if crs is not None:
             try:
                 crs = CRS.from_user_input(crs)
             except CRSError:
                 raise CRSError(f'crs "{crs}" is not a valid Coordinate Reference System')
-            assert (crs.is_geographic == 0), "Provided crs must be projected with units like [m]"
+            assert (crs.is_geographic == 0), "Prodstvided crs must be projected with units like [m]"
             self.crs = crs.to_wkt()
         if resolution is not None:
             self.resolution = resolution
         if lens_position is not None:
             self.set_lens_position(*lens_position)
+        else:
+            self.lens_position = None
         if gcps is not None:
             self.set_gcps(**gcps)
-        if lens_pars is not None:
-            self.set_lens_pars(**lens_pars)
+        if camera_matrix is None or dist_coeffs is None:
+            if self.is_nadir:
+                # with nadir, no perspective can be constructed, hence, camera matrix and dist coeffs will be set to default values
+                self.camera_matrix = cv._get_cam_mtx(self.height, self.width)
+                self.dist_coeffs = cv.DIST_COEFFS
+            # camera pars are incomplete and need to be derived
+            else:
+                self.set_intrinsic(
+                    camera_matrix=camera_matrix,
+                    lens_pars=lens_pars
+                )
         else:
-            self.set_lens_pars()
+            # camera matrix and dist coeffs can also be set hard, this overrules the lens_pars option
+            self.camera_matrix = camera_matrix
+            self.dist_coeffs = dist_coeffs
         if calibration_video is not None:
             self.set_lens_calibration(calibration_video, plot=False)
-        # camera matrix and dist coeffs can also be set hard, this overrules the lens_pars option
-        if camera_matrix is not None:
-            self.camera_matrix = camera_matrix
-        if dist_coeffs is not None:
-            self.dist_coeffs = dist_coeffs
         if bbox is not None:
             self.bbox = bbox
         if window_size is not None:
@@ -117,6 +140,10 @@ class CameraConfig:
         # override the transform and bbox with the set corners
         if corners is not None:
             self.set_bbox_from_corners(corners)
+        if stabilize is not None:
+            self.stabilize = stabilize
+        if rotation is not None:
+            self.rotation = rotation
 
     @property
     def bbox(self):
@@ -133,7 +160,10 @@ class CameraConfig:
 
     @bbox.setter
     def bbox(self, pol):
-        self._bbox = pol
+        if isinstance(pol, str):
+            self._bbox = wkt.loads(pol)
+        else:
+            self._bbox = pol
 
     @property
     def camera_matrix(self):
@@ -151,32 +181,71 @@ class CameraConfig:
     def dist_coeffs(self, dist_coeffs):
         self._dist_coeffs = dist_coeffs.tolist() if isinstance(dist_coeffs, np.ndarray) else dist_coeffs
 
+
     @property
-    def gcp_reduced(self):
+    def gcps_dest(self):
+        """
+
+        Returns
+        -------
+        dst : np.ndarray
+            destination coordinates of ground control point. z-coordinates are parsed from z_0 if necessary
+        """
+        if hasattr(self, "gcps"):
+            if "dst" in self.gcps:
+                return np.array(self.gcps["dst"] if len(self.gcps["dst"][0]) == 3 else np.c_[self.gcps["dst"], np.ones(4)*self.gcps["z_0"]])
+        # if conditions are not yet met, then return None
+        return None
+
+    @property
+    def gcps_dest_bbox(self):
+        """
+
+        Returns
+        -------
+        dst : np.ndarray
+            Destination coordinates measured as column, row in the intended bounding box with the intended resolution
+        """
+        return np.array(cv.transform_to_bbox(self.gcps_dest, self.bbox, self.resolution))
+
+
+    @property
+    def gcps_bbox_reduced(self):
+        """
+
+        Returns
+        -------
+        dst : np.ndarray
+            Destination coordinates in col, row in the intended bounding box, reduced with their mean coordinate
+
+        """
+        return self.gcps_dest_bbox - self.gcps_dest_bbox.mean(axis=0)
+    @property
+    def gcps_reduced(self):
         """
         Get the location of gcp destination points, reduced with their mean for better local projection
 
         Returns
         -------
-        gcp_reduced : np.ndarray
+        dst : np.ndarray
             Reduced coordinate (x, y) or (x, y, z) of gcp destination points
         """
-        return np.array(self.gcps["dst"]) - self.gcp_mean
+        return np.array(self.gcps_dest - self.gcps_mean)
 
     @property
-    def gcp_mean(self):
+    def gcps_mean(self):
         """
         Get the mean location of gcp destination points
 
         Returns
         -------
-        gcp_mean : np.ndarray
+        dst_mean : np.ndarray
             mean coordinate (x, y) or (x, y, z) of gcp destination points
         """
-        return np.array(self.gcps["dst"]).mean(axis=0)
+        return np.array(self.gcps_dest).mean(axis=0)
 
     @property
-    def gcp_dims(self):
+    def gcps_dims(self):
         """
 
         Returns
@@ -185,7 +254,37 @@ class CameraConfig:
             amount of dimensions of gcps (can be 2 or 3)
 
         """
-        return len(self.gcps["dst"][0])
+        return len(self.gcps["dst"][0]) if hasattr(self, "gcps") else None
+
+    @property
+    def is_nadir(self):
+        """
+        Returns if the camera configuration belongs to nadir video
+
+        Returns
+        -------
+        is_nadir : bool
+            False if not nadir, True if nadir
+
+        """
+        return self._is_nadir
+
+    @is_nadir.setter
+    def is_nadir(
+            self,
+            nadir_prop: bool
+    ):
+        self._is_nadir = nadir_prop
+
+    @property
+    def pnp(self):
+        return cv.solvepnp(
+            self.gcps_reduced,
+            self.gcps["src"],
+            self.camera_matrix,
+            self.dist_coeffs
+        )
+
 
     @property
     def shape(self):
@@ -206,6 +305,49 @@ class CameraConfig:
         )
         return rows, cols
 
+    @property
+    def stabilize(self):
+        """
+        Return stabilization polygon (anything outside is used for stabilization
+
+        Returns
+        -------
+        coords : list of lists
+            coordinates in original image frame comprising the polygon for use in stabilization
+        """
+        return self._stabilize
+
+    @stabilize.setter
+    def stabilize(
+            self,
+            coords: List[List[float]]
+    ):
+        self._stabilize = coords
+
+
+    @property
+    def rotation(self):
+        """
+        Return rotation OpenCV code
+
+        Returns
+        -------
+        code : int
+            integer code belonging to rotation, 0 for 90 deg, 1 for 180 deg and 2 for 270 deg
+        """
+        if hasattr(self, "_rotation"):
+            return self._rotation
+        else:
+            return None
+
+    @rotation.setter
+    def rotation(
+            self,
+            rotation_code: int
+    ):
+        self._rotation = rotation_code
+
+
     @property
     def transform(self):
         """
@@ -220,11 +362,11 @@ class CameraConfig:
 
     def set_lens_calibration(
         self,
-        fn,
-        chessboard_size=(9, 6),
-        max_imgs=30,
-        plot=True,
-        progress_bar=True,
+        fn: str,
+        chessboard_size: Optional[Tuple] = (9, 6),
+        max_imgs: Optional[int] = 30,
+        plot: Optional[bool] = True,
+        progress_bar: Optional[bool] = True,
         **kwargs
     ):
         """
@@ -265,7 +407,13 @@ class CameraConfig:
         self.camera_matrix = camera_matrix
         self.dist_coeffs = dist_coeffs
 
-    def get_bbox(self, camera=False, h_a=None, redistort=False):
+    def get_bbox(
+            self,
+            camera: Optional[bool] = False,
+            h_a: Optional[float] = None,
+            within_image: Optional[bool] = False,
+            expand_exterior=True,
+    ) -> Polygon:
         """
 
         Parameters
@@ -277,10 +425,10 @@ class CameraConfig:
             If set with ``camera=True``, then the bbox coordinates will be transformed to the camera perspective,
             using h_a as a present water level. In case a video with higher (lower) water levels is used, this
             will result in a different perspective plane than the control video.
-        redistort : bool, optional
-            If set in combination with ``camera``, the bbox will be redistorted in the camera objective using the
-            distortion coefficients and camera matrix. Not used in orthorectification because this occurs by default
-            on already undistorted images.
+        within_image : bool, optional (default False)
+            Set to True to make an attempt to remove parts of the polygon that lie outside of the image field of view
+        expand_exterior : bool, optional
+            Set to True to expand the corner points to more points. This is particularly useful for plotting purposes.
 
         Returns
         -------
@@ -292,23 +440,63 @@ class CameraConfig:
         This can then be used to reconstruct the grid for velocimetry calculations.
         """
         bbox = self.bbox
-        if camera:
-            coords = np.array(bbox.exterior.coords)
-            # convert to perspective rowcol coordinates
-            M = self.get_M(reverse=True, h_a=h_a)
-            # reduce coords by control point mean
-            coords -= self.gcp_mean[0:2]
-            # TODO: re-distort if needed
-            corners = cv2.perspectiveTransform(np.float32([coords]), M)[0]
-            if redistort:
-                # for visualization on still distorted frames this can be done. DO NOT do this if used for
-                # orthorectification, as this typically occurs on undistorted images.
-                corners = cv.undistort_points(corners, self.camera_matrix, self.dist_coeffs, reverse=True)
+        coords = np.array(bbox.exterior.coords)
+        if within_image:
+            # in this case, always more points than just corners are needed, so expand_exterior is forced to True
+            expand_exterior = True
+        if expand_exterior:
+            # make a new set of bbox coordinates with a higher density. This is meant to enable plotting of distortion on
+            # image frame, and to plot partial coverage in the real-world coordinates
+            coords_expand = np.zeros((0, 2))
+            for n in range(0, len(coords)-1):
+                new_coords = np.linspace(coords[n], coords[n + 1], 100)
+                coords_expand = np.r_[coords_expand, new_coords]
+            coords = coords_expand
+        z_a = self.get_z_a(h_a)
+        # add vertical coordinates to the set
+        coords = np.c_[coords, np.ones(len(coords))*z_a]
+        # project points to pixel image coordinates
+        corners = self.project_points(coords, within_image=within_image)
+        corners = corners[np.isfinite(corners[:, 0])]
+        if not(camera):
+            # project back to real-world coordinates after possibly cutting at edges of visibility
+            corners = self.unproject_points(np.array(np.array(list(zip(*corners))).T), z_a)
+            # corners = self.unproject_points(list(zip(coords[:, 0], coords[:, 1])), z_a)
+            # project points (after cutting on image edges) to geographical space
+        bbox = Polygon(corners[np.isfinite(corners[:, 0])][:, 0:2])
 
-            bbox = Polygon(corners)
         return bbox
 
-    def get_depth(self, z, h_a=None):
+    def get_camera_coords(
+            self,
+            points: List[List],
+    ):
+        """
+        Convert real-world coordinates into camera coordinates using
+        Parameters
+        ----------
+        points : array-like list (of lists)
+            [x, y, z] real-world coordinates
+
+        Returns
+        -------
+        cam_points : np.ndarray (of points)
+            [x, y, z] camera coordinates
+        """
+        _, rvec, tvec = self.pnp
+        # get rotation matrix
+        R, _ = cv2.Rodrigues(rvec)
+
+        # convert points into array
+        points = np.array(points)
+        cam_points = np.einsum('ij,kj->ki', R, np.array(points)) + tvec.flatten()
+        return cam_points
+
+    def get_depth(
+            self,
+            z: List[float],
+            h_a: Optional[float] = None
+    ) -> List[float]:
         """
         Retrieve depth for measured bathymetry points using the camera configuration and an actual water level,
         measured in local reference (e.g. staff gauge).
@@ -331,34 +519,14 @@ class CameraConfig:
         z_pressure = np.maximum(self.gcps["z_0"] - self.gcps["h_ref"] + h_a, z)
         return z_pressure - z
 
-    def get_dst_a(self, h_a=None):
-        """
-        h_a : float, optional
-            actual water level measured [m], if not set, assumption is that a single video
-            is processed and thus changes in water level are not relevant. (default: None)
 
-        Returns
-        -------
-        Actual locations of control points (in case these are only x, y) given the current set water level and
-        the camera location
-        """
-        # map where the destination points are with the actual water level h_a.
-        if h_a is None or h_a == self.gcps["h_ref"]:
-            # fill in the same value for h_ref and h_a
-            dst_a = self.gcps["dst"]
-        else:
-            h_ref = self.gcps["h_ref"]
-            lens_position = self.lens_position
-            dst_a = cv._get_gcps_a(
-                lens_position,
-                h_a,
-                self.gcps["dst"],
-                self.gcps["z_0"],
-                h_ref,
-            )
-        return dst_a
-
-    def get_dist_shore(self, x, y, z, h_a=None):
+    def get_dist_shore(
+            self,
+            x: List[float],
+            y: List[float],
+            z: List[float],
+            h_a: Optional[float] = None
+    ) -> List[float]:
         """
         Retrieve depth for measured bathymetry points using the camera configuration and an actual water level, measured
         in local reference (e.g. staff gauge).
@@ -395,13 +563,43 @@ class CameraConfig:
         dist_shore = np.array([(((x[z_dry] - _x) ** 2 + (y[z_dry] - _y) ** 2) ** 0.5).min() for _x, _y, in zip(x, y)])
         return dist_shore
 
-    def get_dist_wall(self, x, y, z, h_a=None):
+    def get_dist_wall(
+            self,
+            x: List[float],
+            y: List[float],
+            z: List[float],
+            h_a: Optional[float] = None
+    ) -> List[float]:
+        """
+        Retrieve distance to wall for measured bathymetry points using the camera configuration and an actual water
+        level, measured in local reference (e.g. staff gauge).
+
+        Parameters
+        ----------
+        x : list of floats
+            measured bathymetry point x-coordinates
+        y : list of floats
+            measured bathymetry point y-coordinates
+        z : list of floats
+            measured bathymetry point depths
+        h_a : float, optional
+            actual water level measured [m], if not set, assumption is that a single video
+            is processed and thus changes in water level are not relevant. (default: None)
+
+        Returns
+        -------
+        distance : list of floats
+
+        """
         depth = self.get_depth(z, h_a=h_a)
         dist_shore = self.get_dist_shore(x, y, z, h_a=h_a)
         dist_wall = (dist_shore**2 + depth**2)**0.5
         return dist_wall
 
-    def z_to_h(self, z):
+    def z_to_h(
+            self,
+            z: float
+    ) -> float:
         """Convert z coordinates of bathymetry to height coordinates in local reference (e.g. staff gauge)
 
         Parameters
@@ -417,7 +615,12 @@ class CameraConfig:
         h = z + h_ref - self.gcps["z_0"]
         return h
 
-    def get_M(self, h_a=None, to_bbox_grid=False, reverse=False):
+    def get_M(
+            self,
+            h_a: Optional[float] = None,
+            to_bbox_grid: Optional[bool] = False,
+            reverse: Optional[bool] = False
+    ) -> np.ndarray:
         """Establish a transformation matrix for a certain actual water level `h_a`. This is done by mapping where the
         ground control points, measured at `h_ref` will end up with new water level `h_a`, given the lens position.
 
@@ -439,42 +642,28 @@ class CameraConfig:
         """
         src = cv.undistort_points(self.gcps["src"], self.camera_matrix, self.dist_coeffs)
         if to_bbox_grid:
-            # lookup where the destination points are in row/column space
-            # dst_a is the destination point locations position with the actual water level
-            dst_a = cv.transform_to_bbox(
-                self.get_dst_a(h_a),
-                self.bbox,
-                self.resolution
-            )
-            dst_a = np.array(dst_a)
+            dst_a = self.gcps_bbox_reduced
         else:
-            # in case we are dealing with a 2D 4-point, then reproject points on water surface, else keep 3D points
-            dst_a = self.get_dst_a(h_a) if self.gcp_dims == 2 else self.gcps["dst"]
-            # reduce dst_a with its mean to get much more accurate projection result in case x and y order of
-            # magnitude is very large
-            dst_a -= self.gcp_mean
-        # src_a = self.get_src(**lens_pars)
-        if dst_a.shape[-1] == 3:
-            # compute the water level in the coordinate system reduced with the mean gcp coordinate
-            z_a = self.get_z_a(h_a)
-            z_a -= self.gcp_mean[-1]
-            # treating 3D homography
-            return cv.get_M_3D(
-                src=src,
-                dst=dst_a,
-                camera_matrix=self.camera_matrix,
-                dist_coeffs=self.dist_coeffs,
-                z=z_a,
-                reverse=reverse
-            )
-        else:
-            return cv.get_M_2D(
-                src=src,
-                dst=dst_a,
-                reverse=reverse
-            )
+            dst_a = self.gcps_reduced
+        # compute the water level in the coordinate system reduced with the mean gcp coordinate
+        z_a = self.get_z_a(h_a)
+        z_a -= self.gcps_mean[-1]
+        # treating 3D homography
+        print(dst_a)
+        # print(z_a)
+        return cv.get_M_3D(
+            src=src,
+            dst=dst_a,
+            camera_matrix=self.camera_matrix,
+            dist_coeffs=cv.DIST_COEFFS, # self.dist_coeffs,
+            z=z_a,
+            reverse=reverse
+        )
 
-    def get_z_a(self, h_a=None):
+    def get_z_a(
+            self,
+            h_a: Optional[float] = None
+    ) -> float:
         """
         h_a : float, optional
             actual water level measured [m], if not set, assumption is that a single video
@@ -490,7 +679,10 @@ class CameraConfig:
         else:
             return self.gcps["z_0"] + (h_a - self.gcps["h_ref"])
 
-    def set_bbox_from_corners(self, corners):
+    def set_bbox_from_corners(
+            self,
+            corners: List[List[float]]
+    ):
         """
         Establish bbox based on a set of camera perspective corner points Assign corner coordinates to camera
         configuration
@@ -498,26 +690,56 @@ class CameraConfig:
         Parameters
         ----------
         corners : list of lists (4)
-            [columns, row] coordinates in camera perspective
+            [columns, row] coordinates in original camera perspective without any undistortion applied
 
         """
         assert (np.array(corners).shape == (4,
                                             2)), f"a list of lists of 4 coordinates must be given, resulting in (4, " \
                                                  f"2) shape. Current shape is {corners.shape} "
 
-        # get homography, this is the only place besides self.get_M where cv.get_M is used.
-        M_gcp = self.get_M()
-        # TODO: make derivation dependent on 3D or 2D point availability
-        # if self.gcps["src"].shape == 3:
-        #     # TODO: homography from solvepnp
-        bbox = cv.get_aoi(M_gcp, corners, resolution=self.resolution)
-        # bbox is offset by self.gcp_mean. Regenerate bbox after adding offset
-        bbox_xy = np.array(bbox.exterior.coords)
-        bbox_xy += self.gcp_mean[0:2]
-        bbox = Polygon(bbox_xy)
+        # get homography
+        corners_xyz = self.unproject_points(corners, np.ones(4)*self.gcps["z_0"])
+        bbox = cv.get_aoi(
+            corners_xyz,
+            resolution=self.resolution
+        )
         self.bbox = bbox
 
-    def set_lens_pars(self, k1=0, c=2, f=4):
+
+    def set_intrinsic(
+            self,
+            camera_matrix: Optional[List[List]] = None,
+            dist_coeffs: Optional[List[List]] = None,
+            lens_pars: Optional[Dict[str, float]] = None
+    ):
+        # first set a default estimate from pose if 3D gcps are available
+        self.set_lens_pars()  # default parameters use width of frame
+        if hasattr(self, "gcps"):
+            if len(self.gcps["src"]) >= 4:
+            # if self.gcp_dims == 3:
+                self.camera_matrix, self.dist_coeffs, err = cv.optimize_intrinsic(
+                    self.gcps["src"],
+                    self.gcps_dest,
+                    # self.gcps["dst"],
+                    self.height,
+                    self.width,
+                    lens_position=self.lens_position
+                )
+            if lens_pars is not None:
+                # override with lens parameter set by user
+                self.set_lens_pars(**lens_pars)
+            if camera_matrix is not None and dist_coeffs is not None:
+                # override with
+                self.camera_matrix = camera_matrix
+                self.dist_coeffs = dist_coeffs
+
+
+    def set_lens_pars(
+            self,
+            k1: Optional[float] = 0.,
+            c: Optional[float] = 2.,
+            focal_length: Optional[float] = None
+    ):
         """Set the lens parameters of the given CameraConfig
 
         Parameters
@@ -533,20 +755,28 @@ class CameraConfig:
 
         """
         assert (isinstance(k1, (int, float))), "k1 must be a float"
-        assert (isinstance(c, (int, float))), "k1 must be a float"
-        assert (isinstance(f, (int, float))), "k1 must be a float"
+        assert (isinstance(c, (int, float))), "c must be a float"
+        if focal_length is not None:
+            assert (isinstance(focal_length, (int, float, None))), "f must be a float"
         self.dist_coeffs = cv._get_dist_coefs(k1)
-        self.camera_matrix = cv._get_cam_mtx(self.height, self.width, c=c, f=f)
+        self.camera_matrix = cv._get_cam_mtx(self.height, self.width, c=c, focal_length=focal_length)
 
-    def set_gcps(self, src, dst, z_0, h_ref=None, crs=None):
+    def set_gcps(
+            self,
+            src: List[List],
+            dst: List[List],
+            z_0: float,
+            h_ref: Optional[float] = None,
+            crs: Optional[Any] = None
+    ):
         """
         Set ground control points for the given CameraConfig
 
         Parameters
         ----------
-        src : list of lists (4 or 6+)
+        src : list of lists (2, 4 or 6+)
             [x, y] pairs of columns and rows in the frames of the original video
-        dst : list of lists (4 or 6+)
+        dst : list of lists (2, 4 or 6+)
             [x, y] or [x, y, z] pairs of real world coordinates in the given coordinate reference system.
         z_0 : float
             Water level measured in global reference system such as a geoid or ellipsoid used
@@ -566,8 +796,11 @@ class CameraConfig:
         assert (isinstance(src, list)), f"src must be a list of (x, y) or (x, y, z) coordinates"
         assert (isinstance(dst, list)), f"dst must be a list of (x, y) or (x, y, z) coordinates"
         if np.array(dst).shape[1] == 2:
-            assert (len(src) == 4), f"4 source points are expected in src, but {len(src)} were found"
-            assert (len(dst) == 4), f"4 destination points are expected in dst, but {len(dst)} were found"
+            assert (len(src) in [2, 4]), f"2 or 4 source points are expected in src, but {len(src)} were found"
+            if len(src) == 4:
+                assert (len(dst) == 4), f"4 destination points are expected in dst, but {len(dst)} were found"
+            else:
+                assert (len(dst) == 2), f"2 destination points are expected in dst, but {len(dst)} were found"
         else:
             assert(len(src) == len(dst)), f"Amount of (x, y, z) coordinates in src ({len(src)}) and dst ({len(dst)} must be equal"
             assert(len(src) >= 6), f"for (x, y, z) points, at least 6 pairs must be available, only {len(src)} provided"
@@ -585,6 +818,10 @@ class CameraConfig:
                     'with a spirit level.')
             dst = helpers.xyz_transform(dst, crs, CRS.from_wkt(self.crs))
         # if there is no h_ref, then no local gauge system, so set h_ref to zero
+        # check if 2 points are available
+        if len(src) == 2:
+            self.is_nadir = True
+            src, dst = cv._get_gcps_2_4(src, dst, self.width, self.height)
         if h_ref is None:
             h_ref = 0.
         self.gcps = {
@@ -594,7 +831,13 @@ class CameraConfig:
             "z_0": z_0,
         }
 
-    def set_lens_position(self, x, y, z, crs=None):
+    def set_lens_position(
+            self,
+            x: float,
+            y: float,
+            z: float,
+            crs: Optional[Any] = None
+    ):
         """Set the geographical position of the lens of current CameraConfig.
 
         Parameters
@@ -617,16 +860,103 @@ class CameraConfig:
             x, y = helpers.xyz_transform([[x, y]], crs, self.crs)[0]
         self.lens_position = [x, y, z]
 
+    def project_points(
+            self,
+            points: List[List],
+            within_image=False
+    ) -> np.ndarray:
+        """
+        Project real world x, y, z coordinates into col, row coordinates on image. If
+        col, row coordinates are not allowed to go outside of the image frame, then set within_image = True
+
+        Parameters
+        ----------
+        points : list of lists or array-like
+            list of points [x, y, z] in real world coordinates
+
+        Returns
+        -------
+        points_project : list or array-like
+            list of points (equal in length as points) with [col, row] coordinates
+        """
+        _, rvec, tvec = self.pnp
+        # normalize points wrt mean of gcps
+        points = np.float32(np.array(points) - self.gcps_mean)
+        points_proj, jacobian = cv2.projectPoints(
+            points,
+            rvec,
+            tvec,
+            np.array(self.camera_matrix),
+            np.array(self.dist_coeffs)
+        )
+        points_proj = np.array([list(point[0]) for point in points_proj])
+
+        # points_back = cv.unproject_points(src=points_proj, z=points[:, -1], )
+        if within_image:
+            # back project, the ones that do not get close to the original will be removed
+            points_back = cv.unproject_points(
+                points_proj,
+                z=points[:, -1],
+                rvec=rvec,
+                tvec=tvec,
+                camera_matrix=self.camera_matrix,
+                dist_coeffs=self.dist_coeffs
+            )
+            # TODO: figure out how to filter out points that bend into the image frame according to the
+            # distortion parameters, but are in fact outside. This is not yet working the way it should
+            filter = np.all(np.isclose(np.array(points_back), np.array(points), atol=1e-2), axis=1)
+            points_proj[~filter] = np.nan
+            # also filter points outside edges of image
+            points_proj[points_proj[:, 0] < 0, 0] = 0.
+            points_proj[points_proj[:, 0] > self.width - 1, 0] = self.width - 1
+            points_proj[points_proj[:, 1] < 0, 1] = 0.
+            points_proj[points_proj[:, 1] > self.height - 1, 1] = self.height - 1
+        return points_proj
+
+
+    def unproject_points(
+            self,
+            points: List[List],
+            zs: Union[float, List[float]]
+    ) -> np.ndarray:
+        """
+        Reverse projects points in [column, row] space to [x, y, z] real world
+        Parameters
+        ----------
+        points : List of lists or array-like
+            Points in [col, row] to unproject
+        zs : float or list of floats : z-coordinate on which to unproject points
+
+        Returns
+        -------
+        points_unproject : List of lists or array-like
+            unprojected points as list of [x, y, z] coordinates
+        """
+        _, rvec, tvec = self.pnp
+        # reduce zs by the mean of the gcps
+        _zs = zs - self.gcps_mean[-1]
+        dst = cv.unproject_points(
+            np.array(points),
+            _zs,
+            rvec=rvec,
+            tvec=tvec,
+            camera_matrix=self.camera_matrix,
+            dist_coeffs=self.dist_coeffs
+        )
+        dst = np.array(dst) + self.gcps_mean
+        return dst
+
+
     def plot(
             self,
-            figsize=(13, 8),
-            ax=None,
-            tiles=None,
-            buffer=0.0005,
-            zoom_level=19,
-            camera=False,
-            tiles_kwargs={}
-    ):
+            figsize: Optional[Tuple] = (13, 8),
+            ax: Optional[plt.Axes] = None,
+            tiles: Optional[Any] = None,
+            buffer: Optional[float] = 0.0005,
+            zoom_level: Optional[int] = 19,
+            camera: Optional[bool] = False,
+            tiles_kwargs: Optional[Dict] = {}
+    ) -> plt.Axes:
         """
         Plot the geographical situation of the CameraConfig. This is very useful to check if the CameraConfig seems
         to be in the right location. Requires cartopy to be installed.
@@ -668,7 +998,9 @@ class CameraConfig:
             points = [Point(p[0], p[1]) for p in self.gcps["dst"]]
 
         if not camera:
-            if hasattr(self, "lens_position") and not camera:
+            if self.lens_position is not None and not camera:
+            #
+            # if hasattr(self, "lens_position") and not camera:
                 points.append(Point(self.lens_position[0], self.lens_position[1]))
             # transform points in case a crs is provided
             if hasattr(self, "crs"):
@@ -724,7 +1056,7 @@ class CameraConfig:
             **plot_kwargs
         }
         if hasattr(self, "bbox"):
-            self.plot_bbox(ax=ax, camera=camera, transformer=transformer, **patch_kwargs)
+            self.plot_bbox(ax=ax, camera=camera, transformer=transformer, within_image=True, **patch_kwargs)
         if camera:
             # make sure that zero is on the top
             ax.set_aspect("equal")
@@ -734,7 +1066,15 @@ class CameraConfig:
         ax.legend()
         return ax
 
-    def plot_bbox(self, ax=None, camera=False, transformer=None, h_a=None, **kwargs):
+    def plot_bbox(
+            self,
+            ax: Optional[plt.Axes] = None,
+            camera: Optional[bool] = False,
+            transformer: Optional[Any] = None,
+            h_a: Optional[float] = None,
+            within_image: Optional[bool] = True,
+            **kwargs
+    ):
         """
         Plot bounding box for orthorectification in a geographical projection (``camera=False``) or the camera
         Field Of View (``camera=True``).
@@ -757,7 +1097,7 @@ class CameraConfig:
         p : matplotlib.patch mappable
         """
         # collect information to plot
-        bbox = self.get_bbox(camera=camera, h_a=h_a)
+        bbox = self.get_bbox(camera=camera, h_a=h_a, within_image=within_image)
         if camera is False and transformer is not None:
             # geographical projection is needed
             bbox = ops.transform(transformer, bbox)
@@ -771,7 +1111,8 @@ class CameraConfig:
         p = ax.add_patch(patch)
         return p
 
-    def to_dict(self):
+
+    def to_dict(self) -> Dict:
         """Return the CameraConfig object as dictionary
 
         Returns
@@ -789,13 +1130,16 @@ class CameraConfig:
 
         return d
 
-    def to_dict_str(self):
+    def to_dict_str(self) -> Dict:
         d = self.to_dict()
         # convert anything that is not string in string
         dict_str = {k: v if not(isinstance(v, Polygon)) else v.__str__() for k, v in d.items()}
         return dict_str
 
-    def to_file(self, fn):
+    def to_file(
+            self,
+            fn: str
+    ):
         """Write the CameraConfig object to json structure
         
         Parameters
@@ -808,7 +1152,7 @@ class CameraConfig:
         with open(fn, "w") as f:
             f.write(self.to_json())
 
-    def to_json(self):
+    def to_json(self) -> str:
         """Convert CameraConfig object to string
         
         Returns
@@ -832,7 +1176,9 @@ look as follows for a HD video:
 """
 
 
-def get_camera_config(s):
+def get_camera_config(
+        s: str
+) -> CameraConfig:
     """Read camera config from string
 
     Parameters
@@ -855,7 +1201,10 @@ def get_camera_config(s):
     return CameraConfig(**d)
 
 
-def load_camera_config(fn):
+
+def load_camera_config(
+        fn: str
+) -> CameraConfig:
     """Load a CameraConfig from a geojson file.
 
     Parameters
