@@ -1,3 +1,4 @@
+import copy
 import functools
 import matplotlib.pyplot as plt
 import numpy as np
@@ -39,7 +40,19 @@ def _base_plot(plot_func):
 
     # apply wrapper to allow for partial update of the function, with updated docstring
     @functools.wraps(plot_func)
-    def get_plot_method(ref, mode="local", ax=None, add_colorbar=False, add_cross_section=True, kwargs_line={}, *args, **kwargs):
+    def get_plot_method(
+            ref,
+            mode="local",
+            ax=None,
+            add_colorbar=False,
+            add_cross_section=True,
+            add_text=False,
+            text_prefix="",
+            text_suffix="",
+            kwargs_line={},
+            *args,
+            **kwargs
+    ):
         """Retrieve plot method with all required inputs
 
         Parameters
@@ -56,6 +69,12 @@ def _base_plot(plot_func):
             if True, a colorbar is added to axes (default: False)
         add_cross_section : boolean, optional
             if True, and a transect is plotted, the transect coordinates are plotted (default: True)
+        add_text : boolean, optional
+            if True, add a text label in the axes displaying information about the video's transect
+        text_prefix : str, optional
+            string to add in front of standard text on transect plot. Only used if ``add_text=True``
+        text_suffix : str, optional
+            String to add after standard text on transect plot. Only used if ``add_text=True``
         kwargs_line : dict, optional
             additional keyword arguments passed to matplotlib.pyplot.plot for plotting cross-section.
             (Default value = {})
@@ -69,6 +88,11 @@ def _base_plot(plot_func):
 
         """
         # in case persistent transform is in kwargs_line, remove this
+        if mode == "geographical":
+            try:
+                import cartopy
+            except:
+                raise ImportError("Cartopy not found, please install with 'mamba install -c conda-forge cartopy")
         if "transform" in kwargs_line:
             del kwargs_line["transform"]
         ax = _prepare_axes(ax=ax, mode=mode)
@@ -149,6 +173,9 @@ def _base_plot(plot_func):
                         ax.plot(x_bottom, y_bottom, "#0088FF", alpha=0.3, linewidth=3, **kwargs_line)
                         ax.plot(x_bottom, y_bottom, "#0088FF", alpha=0.3, linewidth=2, **kwargs_line)
                         ax.plot(x_bottom, y_bottom, "#0088FF", alpha=0.3, linewidth=1, **kwargs_line)
+                if add_text:
+                    plot_text(ax, ref._obj, text_prefix, text_suffix)
+
         if mode == "geographical" and not(is_transect):
             ax.set_extent(
                 [
@@ -163,6 +190,7 @@ def _base_plot(plot_func):
     return get_plot_method
 
 
+
 def _frames_plot(ref, ax=None, mode="local", *args, **kwargs):
     """Creates QuadMesh plot from a RGB or grayscale frame on a new or existing (if ax is not None) axes
     
@@ -185,10 +213,19 @@ def _frames_plot(ref, ax=None, mode="local", *args, **kwargs):
 
     """
     # prepare axes
+    if mode == "geographical":
+        try:
+            import cartopy
+        except:
+            raise ImportError("Cartopy not found, please install with 'mamba install -c conda-forge cartopy")
     if "time" in ref._obj.coords:
         if ref._obj.time.size > 1:
             raise AttributeError(f'Object contains dimension "time" with length {len(ref._obj.time)}. Reduce dataset by selecting one time step or taking a median, mean or other statistic.')
-    ax = _prepare_axes(ax=ax, mode=mode)
+    if mode == "camera":
+        rotation = ref.camera_config.rotation
+    else:
+        rotation = None
+    ax = _prepare_axes(ax=ax, mode=mode, rotation=rotation)
     f = ax.figure  # handle to figure
     if mode == "local":
         x = "x"
@@ -287,8 +324,9 @@ class _Transect_PlotMethods:
         transect = self._obj.transect
         camera_config = transect.camera_config
         # M = velocimetry.camera_config.get_M(velocimetry.h_a, to_bbox_grid=True, reverse=True)
-        src = camera_config.get_bbox(camera=True, h_a=transect.h_a).exterior.coords[0:4]
-        dst_xy = camera_config.get_bbox().exterior.coords[0:4]
+        src = camera_config.get_bbox(camera=True, h_a=transect.h_a, expand_exterior=False).exterior.coords[0:4]
+        dst_xy = camera_config.get_bbox(expand_exterior=False).exterior.coords[0:4]
+        # get geographic coordinates bbox corners
         # get geographic coordinates bbox corners
         dst = cv.transform_to_bbox(dst_xy, camera_config.bbox, camera_config.resolution)
         M = cv.get_M_2D(src, dst, reverse=True)
@@ -449,8 +487,8 @@ class _Velocimetry_PlotMethods:
         velocimetry = self._obj.velocimetry
         camera_config = velocimetry.camera_config
         # M = velocimetry.camera_config.get_M(velocimetry.h_a, to_bbox_grid=True, reverse=True)
-        src = camera_config.get_bbox(camera=True, h_a=velocimetry.h_a).exterior.coords[0:4]
-        dst_xy = camera_config.get_bbox().exterior.coords[0:4]
+        src = camera_config.get_bbox(camera=True, h_a=velocimetry.h_a, expand_exterior=False).exterior.coords[0:4]
+        dst_xy = camera_config.get_bbox(expand_exterior=False).exterior.coords[0:4]
         # get geographic coordinates bbox corners
         dst = cv.transform_to_bbox(dst_xy, camera_config.bbox, camera_config.resolution)
         M = cv.get_M_2D(src, dst, reverse=True)
@@ -576,12 +614,46 @@ def cbar(ax, p, size=12, **kwargs):
     cb.set_label(label="velocity [m/s]", size=size, path_effects=path_effects)
     return cb
 
+def plot_text(ax, ds, prefix, suffix):
+    if "q" not in ds:
+        return
+    _ds = copy.deepcopy(ds)
+    path_effects = [
+        patheffects.Stroke(linewidth=3, foreground="w"),
+        patheffects.Normal(),
+    ]
+    xloc = 0.95
+    yloc = 0.95
+    _ds.transect.get_river_flow(q_name="q")
+    Q = np.abs(_ds.river_flow)
+    string = prefix
+    string += "Water level: {:1.2f} m\nDischarge: {:1.2f} m3/s".format(_ds.transect.h_a, Q.values)
+    if "q_nofill" in ds:
+        _ds.transect.get_river_flow(q_name="q_nofill")
+        Q_nofill = np.abs(_ds.river_flow)
+        perc_measured = Q_nofill / Q * 100  # fraction that is truly measured compared to total
+        string += " ({:1.0f}% measured)".format(perc_measured.values)
+    # reset river flow if necessary
+    string += suffix
+    ax.text(
+        xloc,
+        yloc,
+        string,
+        size=24,
+        horizontalalignment="right",
+        verticalalignment="top",
+        path_effects=path_effects,
+        transform=ax.transAxes
+    )
 
-def _prepare_axes(ax=None, mode="local"):
+
+def _prepare_axes(ax=None, mode="local", rotation=None):
     """Prepares the axes, needed to plot results, called from `pyorc.PIV.plot`.
 
     Parameters
     ----------
+    ref : Frames, Transect or Velocimetry
+        object to derive plot from
     ax : plt.axes, optional
         if not provided (default), a new axes is prepared (default: None)
     mode : str, optional
@@ -600,8 +672,13 @@ def _prepare_axes(ax=None, mode="local"):
         return ax
 
     # make a screen filling figure with black edges and faces
-    f = plt.figure(figsize=(16, 9), frameon=False, facecolor="k")
-    f.set_size_inches(16, 9, True)
+    if rotation in [90, 270]:
+        f = plt.figure(figsize=(9, 16), frameon=False, facecolor="k")
+        f.set_size_inches(9, 16, True)
+    else:
+        f = plt.figure(figsize=(16, 9), frameon=False, facecolor="k")
+        f.set_size_inches(16, 9, True)
+
     f.patch.set_facecolor("k")
     f.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=None, hspace=None)
     if mode == "geographical":
