@@ -163,11 +163,12 @@ class Transect(ORCBase):
         ----------
         v_corr : float, optional
             correction factor (default: 0.9)
-        fill_method : method to fill missing values. "zeros" fills NaNS with zeros, "interpolate" interpolates values
+        fill_method : str, optional
+            method to fill missing values. "zeros" fills NaNS with zeros, "interpolate" interpolates values
             from nearest neighbour, "log_interp" interpolates values linearly with velocities scaled by the log of
-             depth over a roughness length, "log_fit" fits a 4-parameter logarithmic profile with depth and with
-             changing velocities towards banks on known velocities, and fills missing with the fitted relationship
-             (experimental) (Default value = "zeros").
+            depth over a roughness length, "log_fit" fits a 4-parameter logarithmic profile with depth and with
+            changing velocities towards banks on known velocities, and fills missing with the fitted relationship
+            (experimental) (Default value = "zeros").
 
         Returns
         -------
@@ -185,6 +186,8 @@ class Transect(ORCBase):
         z = ds["zcoords"].values
         # add filled surface velocities with a logarithmic profile curve fit
         depth = self.camera_config.get_depth(z, self.h_a)
+        # make velocities zero where depth is zero
+        ds["v_eff_nofill"][:, depth<=0] = 0.
         if fill_method == "zeros":
             ds["v_eff"] = ds["v_eff_nofill"].fillna(0.)
         elif fill_method == "log_fit":
