@@ -10,11 +10,13 @@ import os
 import warnings
 import xarray as xr
 
-from .. import cv, const
+from typing import List, Optional, Union
+
+from .. import cv, const, helpers
 from .cameraconfig import load_camera_config, get_camera_config, CameraConfig
 
 
-class Video:  #(cv2.VideoCapture)
+class Video:  # (cv2.VideoCapture)
     def __repr__(self):
         template = """
 Filename: {:s}
@@ -34,13 +36,16 @@ Camera configuration: {:s}
 
     def __init__(
             self,
-            fn,
-            camera_config=None,
-            h_a=None,
-            start_frame=None,
-            end_frame=None,
-            stabilize=None,
-            mask_exterior=None,
+            fn: str,
+            camera_config: Optional[Union[str, CameraConfig]] = None,
+            h_a: Optional[float] = None,
+            start_frame: Optional[int] = None,
+            end_frame: Optional[int] = None,
+            freq: Optional[int] = 1,
+            stabilize: Optional[List[List]] = None,
+            lazy: bool = True,
+            rotation: Optional[int] = None,
+            fps: Optional[float] = None,
     ):
         """
         Video class, inheriting parts from cv2.VideoCapture. Contains a camera configuration to it, and a start and end
@@ -62,41 +67,50 @@ Camera configuration: {:s}
             first frame to use in analysis (default: 0)
         end_frame : int, optional
             last frame to use in analysis (if not set, last frame available in video will be used)
-        stabilize : optional
-            If set to a recipe name, the video will be stabilized by attempting to find rigid points and track these with
-            Lukas Kanade optical flow. "fixed" for FOV that is meant to be in one place, "moving" for a moving FOV.
-        mask_exterior : list of lists,
+        stabilize : list of lists, optional
             set of coordinates, that together encapsulate the polygon that defines the mask, separating land from water.
-            The mask is used to select region (on land) for rigid point search for stabilization.
+            The mask is used to select region (on land) for rigid point search for stabilization. If not set, then no
+            stabilization will be performed
+        lazy : bool, optional
+            If set, frames are read lazily. This slows down the processing, but makes interaction with large videos
+            easier and consuming less memory. For operational processing with short videos, it is recommended to set
+            this explicitly to False.
+        rotation : int, optional
+            can be 0, 90, 180, 270. If provided, images will be forced to rotate along the provided angle.
+        fps : float, optional
+            hard set for frames per second. Use this with utmost caution and only when you are confident that the video
+            metadata is incorrect.
         """
-        assert(isinstance(start_frame, (int, type(None)))), 'start_frame must be of type "int"'
-        assert(isinstance(end_frame, (int, type(None)))), 'end_frame must be of type "int"'
-        assert(stabilize in ["fixed", "moving", "all", None]), 'stabilize must be "fixed", "moving" or "all"'
-        self.feats_pos = None
+        assert (isinstance(start_frame, (int, type(None)))), 'start_frame must be of type "int"'
+        assert (isinstance(end_frame, (int, type(None)))), 'end_frame must be of type "int"'
+        # assert (isinstance(stabilize, (list, type(None)))), f'stabilize must contain a list of points, but is {stabilize}'
         self.feats_stats = None
         self.feats_errs = None
         self.ms = None
         self.mask = None
+        self.lazy = lazy
+        self.stabilize = stabilize
         if camera_config is not None:
             self.camera_config = camera_config
-        # if camera_config is not None:
+            # if camera_config is not None:
             # check if h_a is supplied, if so, then also z_0 and h_ref must be available
             if h_a is not None:
-                assert(isinstance(self.camera_config.gcps["z_0"], float)),\
+                assert (isinstance(self.camera_config.gcps["z_0"], float)), \
                     "h_a was supplied, but camera config's gcps do not contain z_0, this is needed for dynamic " \
                     "reprojection. You can supplying z_0 and h_ref in the camera_config's gcps upon making a camera " \
                     "configuration. "
-                assert (isinstance(self.camera_config.gcps["h_ref"], float)),\
+                assert (isinstance(self.camera_config.gcps["h_ref"], float)), \
                     "h_a was supplied, but camera config's gcps do not contain h_ref, this is needed for dynamic " \
                     "reprojection. You must supply z_0 and h_ref in the camera_config's gcps upon making a camera " \
                     "configuration. "
         cap = cv2.VideoCapture(fn)
+        cap.set(cv2.CAP_PROP_ORIENTATION_AUTO, 180.0)
         self.height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
         self.width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
         # explicitly open file for reading
-        if mask_exterior is not None:
-            # set a mask based on the roi points
-            self.set_mask_from_exterior(mask_exterior)
+        if self.stabilize is not None:
+            # set a gridded mask based on the roi points
+            self.set_mask_from_exterior(self.stabilize)
         # set end and start frame
         self.frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
         if start_frame is not None:
@@ -113,33 +127,66 @@ Camera configuration: {:s}
             end_frame = np.minimum(end_frame, self.frame_count)
         else:
             end_frame = self.frame_count
-        # extract times and frame numbers as far as available
-        time, frame_number = cv.get_time_frames(cap, start_frame, end_frame)
+        self.fps = cap.get(cv2.CAP_PROP_FPS)
+        self.rotation = cap.get(cv2.CAP_PROP_ORIENTATION_META)
+        # extract times, frame numbers and frames as far as available
+        time, frame_number, frames = cv.get_time_frames(
+            cap,
+            start_frame,
+            end_frame,
+            lazy=lazy,
+            rotation=self.rotation,
+            method="grayscale",
+            fps=fps
+        )
+        self.frames = frames
         # check if end_frame changed
         if frame_number[-1] != end_frame:
             warnings.warn(f"End frame {end_frame} cannot be read from file. End frame is adapted to {frame_number[-1]}")
             end_frame = frame_number[-1]
 
         self.end_frame = end_frame
+        self.freq = freq
         self.time = time
         self.frame_number = frame_number
         self.start_frame = start_frame
-        if stabilize is not None:
+        if self.stabilize is not None:
             # select the right recipe dependent on the movie being fixed or moving
-            recipe = const.CLASSIFY_CAM[stabilize] if stabilize in const.CLASSIFY_CAM else []
-            self._get_pos_feats(cap, recipe=recipe)
-            self._get_ms()
-
-        self.fps = cap.get(cv2.CAP_PROP_FPS)
+            self.get_ms(cap)
+        if fps is None:
+            self.fps = cap.get(cv2.CAP_PROP_FPS)
+        else:
+            self.fps = fps
+        if rotation is None:
+            rotation = cap.get(cv2.CAP_PROP_ORIENTATION_META)
+            if rotation in [90, 270]:
+                # fix a bug in the cv2 code, that rotates portrait videos in the wrong direction
+                rotation = 180
+        self.rotation = rotation
         # set other properties
         self.h_a = h_a
-        # make camera config part of the vidoe object
+        # make camera config part of the video object
         self.fn = fn
         self._stills = {}  # here all stills are stored lazily
         # nothing to be done at this stage, release file for now.
         cap.release()
         del cap
 
+    @property
+    def lazy(self):
+        """
+
+        Returns
+        -------
+        np.ndarray
+            Mask of region of interest
+        """
+        return self._lazy
+
+    @lazy.setter
+    def lazy(self, lazy):
+        self._lazy = lazy
+
 
     @property
     def mask(self):
@@ -165,7 +212,10 @@ Camera configuration: {:s}
 
         :return: CameraConfig object
         """
-        return self._camera_config
+        if hasattr(self, "_camera_config"):
+            return self._camera_config
+        else:
+            return None
 
     @camera_config.setter
     def camera_config(self, camera_config_input):
@@ -188,7 +238,8 @@ Camera configuration: {:s}
                 # Create CameraConfig from dict
                 self._camera_config = CameraConfig(**camera_config_input)
         except:
-            raise IOError("Could not recognise input as a CameraConfig file, string, dictionary or CameraConfig object.")
+            raise IOError(
+                "Could not recognise input as a CameraConfig file, string, dictionary or CameraConfig object.")
 
     @property
     def end_frame(self):
@@ -206,6 +257,35 @@ Camera configuration: {:s}
         else:
             self._end_frame = min(self.frame_count - 1, end_frame)
 
+    @property
+    def freq(self):
+        """
+
+        Returns
+        -------
+        int: frequency (1 in nth frames to select)
+
+        """
+        return self._freq
+
+    @freq.setter
+    def freq(self, freq=1):
+        self._freq = freq
+
+    @property
+    def stabilize(self):
+        if self._stabilize is not None:
+            return self._stabilize
+        elif hasattr(self, "camera_config"):
+            if hasattr(self.camera_config, "stabilize"):
+                return self.camera_config.stabilize
+
+    @stabilize.setter
+    def stabilize(
+            self,
+            coords: Optional[List[List]] = None
+    ):
+        self._stabilize = coords
 
     @property
     def h_a(self):
@@ -216,11 +296,15 @@ Camera configuration: {:s}
         return self._h_a
 
     @h_a.setter
-    def h_a(self, h_a):
+    def h_a(
+            self,
+            h_a: float
+    ):
         if h_a is not None:
-            assert(isinstance(h_a, float)), f"The actual water level must be a float, you supplied a {type(h_a)}"
+            assert (isinstance(h_a, float)), f"The actual water level must be a float, you supplied a {type(h_a)}"
             if h_a < 0:
-                warnings.warn("Water level is negative. This can be correct, but may be unlikely, especially if you use a staff gauge.")
+                warnings.warn(
+                    "Water level is negative. This can be correct, but may be unlikely, especially if you use a staff gauge.")
         self._h_a = h_a
 
     @property
@@ -232,12 +316,27 @@ Camera configuration: {:s}
         return self._start_frame
 
     @start_frame.setter
-    def start_frame(self, start_frame=None):
+    def start_frame(
+            self,
+            start_frame: Optional[int] = None
+    ):
         if start_frame is None:
             self._start_frame = 0
         else:
             self._start_frame = start_frame
 
+
+    @property
+    def frames(self):
+        return self._frames
+
+    @frames.setter
+    def frames(
+            self,
+            frames: Optional[List] = None
+    ):
+        self._frames = frames
+
     @property
     def fps(self):
         """
@@ -247,7 +346,10 @@ Camera configuration: {:s}
         return self._fps
 
     @fps.setter
-    def fps(self, fps):
+    def fps(
+            self,
+            fps: float
+    ):
         if (np.isinf(fps)) or (fps <= 0):
             raise ValueError(f"FPS in video is {fps} which is not a valid value. Repair the video file before use")
         self._fps = fps
@@ -261,93 +363,125 @@ Camera configuration: {:s}
         return self._corners
 
     @corners.setter
-    def corners(self, corners):
+    def corners(
+            self,
+            corners: List[List]
+    ):
         self._corners = corners
 
-    def get_frame(self, n, method="grayscale", lens_corr=False):
-        """
-        Retrieve one frame. Frame will be corrected for lens distortion if lens parameters are given.
 
-        :param n: int, frame number to retrieve
-        :param method: str, can be "rgb", "grayscale", or "hsv", default: "grayscale"
-        :param lens_corr: bool, optional, if set to True, lens parameters will be used to undistort image
-        :return: np.ndarray containing frame
+    @property
+    def rotation(self):
+        if self._rotation is not None:
+            return self._rotation
+        elif hasattr(self, "camera_config"):
+            if hasattr(self.camera_config, "rotation"):
+                return helpers.get_rotation_code(self.camera_config.rotation)
+
+
+    @rotation.setter
+    def rotation(
+            self,
+            rotation_code: int
+    ):
         """
-        assert(n >= 0), "frame number cannot be negative"
-        assert(n - self.start_frame <= self.end_frame - self.start_frame), "frame number is larger than the different between the start and end frame"
-        assert(method in ["grayscale", "rgb", "hsv"]), f'method must be "grayscale", "rgb" or "hsv", method is "{method}"'
+        Set rotation from integer in the form of OpenCV rotation codes
+        """
+        self._rotation = helpers.get_rotation_code(rotation_code)
+
+
+    def get_frame(
+            self,
+            n: int,
+            method: Optional[str] = "grayscale",
+    ) -> np.ndarray:
+        """
+        Retrieve one frame.
+
+        Parameters:
+        -----------
+        n : int
+            frame number to retrieve
+        method : str
+            can be "rgb", "grayscale", or "hsv", default: "grayscale"
+
+        Returns
+        -------
+        frame : np.ndarray
+            2d array (grayscale) or 3d (rgb/hsv) with frame
+        """
+        assert (n >= 0), "frame number cannot be negative"
+        assert (
+                n - self.start_frame <= self.end_frame - self.start_frame), "frame number is larger than the different between the start and end frame"
+        assert (method in ["grayscale", "rgb",
+                           "hsv"]), f'method must be "grayscale", "rgb" or "hsv", method is "{method}"'
         cap = cv2.VideoCapture(self.fn)
         cap.set(cv2.CAP_PROP_POS_FRAMES, n + self.start_frame)
-        try:
-            ret, img = cap.read()
-        except:
-            raise IOError(f"Cannot read")
-        if ret:
-            if self.ms is not None:
-                # correct for stabilization
-                h = img.shape[0]
-                w = img.shape[1]
-                img = cv2.warpAffine(img, self.ms[n], (w, h))
-
-            # if lens_corr:
-            #     if self.camera_config.lens_pars is not None:
-            # apply lens distortion correction
-            if hasattr(self, "camera_config"):
-                img = cv.undistort_img(img, self.camera_config.camera_matrix, self.camera_config.dist_coeffs)
-            if method == "grayscale":
-                # apply gray scaling, contrast- and gamma correction
-                # img = _corr_color(img, alpha=None, beta=None, gamma=0.4)
-                img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY) #  mean(axis=2)
-            elif method == "rgb":
-                # turn bgr to rgb for plotting purposes
-                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
-            elif method == "hsv":
-                img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
+        ret, img = cv.get_frame(
+            cap,
+            rotation=self.rotation,
+            ms=self.ms[n] if self.ms else None,
+            method=method
+        )
         self.frame_count = n + 1
         cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
         cap.release()
         return img
 
-    def get_frames(self, **kwargs):
+    def get_frames(
+            self,
+            **kwargs
+    ) -> xr.DataArray:
         """
         Get a xr.DataArray, containing a dask array of frames, from `start_frame` until `end_frame`, expected to be read
         lazily. The xr.DataArray will contain all coordinate variables and attributes, needed for further processing
         steps.
 
-        :param kwargs: dict, keyword arguments to pass to `get_frame`. Currently only `grayscale` is supported.
-        :return: xr.DataArray, containing all requested frames
+        Parameters
+        ----------
+        **kwargs: dict, optional
+            keyword arguments to pass to `get_frame`. Currently only `grayscale` is supported.
+
+        Returns
+        -------
+        frames : xr.DataArray
+            containing all requested frames
         """
-        assert(hasattr(self, "_camera_config")), "No camera configuration is set, add it to the video using the .camera_config method"
+        assert (hasattr(self,
+                        "_camera_config")), "No camera configuration is set, add it to the video using the .camera_config method"
         # camera_config may be altered for the frames object, so copy below
         camera_config = copy.deepcopy(self.camera_config)
-        get_frame = dask.delayed(self.get_frame, pure=True)  # Lazy version of get_frame
-        # get all listed frames
-        frames = [get_frame(n=n, **kwargs) for n, f_number in enumerate(self.frame_number)]
-        sample = frames[0].compute()
-        data_array = [da.from_delayed(
-            frame,
-            dtype=sample.dtype,
-            shape=sample.shape
-        ) for frame in frames]
-        # if "lens_corr" in kwargs:
-        #     if kwargs["lens_corr"]:
-                # also correct the control point src
-        if hasattr(camera_config, "gcps"):
-            camera_config.gcps["src"] = cv.undistort_points(
-                camera_config.gcps["src"],
-                camera_config.camera_matrix,
-                camera_config.dist_coeffs,
-            )
-                # camera_config.corners = cv.undistort_points(
-                #     camera_config.corners,
-                #     sample.shape[0],
-                #     sample.shape[1],
-                #     **self.camera_config.lens_pars
-                # )
-        time = np.array(self.time) * 0.001 # measure in seconds to comply with CF conventions # np.arange(len(data_array))*1/self.fps
+
+        if self.frames is None or len(kwargs) > 0:
+            # a specific method for collecting frames is requested or lazy access is requested.
+            get_frame = dask.delayed(self.get_frame, pure=True)  # Lazy version of get_frame
+            # get all listed frames
+            frames = [get_frame(n=n, **kwargs) for n, f_number in enumerate(self.frame_number)]
+            sample = frames[0].compute()
+            data_array = [da.from_delayed(
+                frame,
+                dtype=sample.dtype,
+                shape=sample.shape
+            ) for frame in frames]
+            da_stack = da.stack(data_array, axis=0)
+        else:
+            sample = self.frames[0]
+            da_stack = self.frames
+
+        # undistort source control points
+        # if hasattr(camera_config, "gcps"):
+        #     camera_config.gcps["src"] = cv.undistort_points(
+        #         camera_config.gcps["src"],
+        #         camera_config.camera_matrix,
+        #         camera_config.dist_coeffs,
+        #     )
+        time = np.array(
+            self.time) * 0.001  # measure in seconds to comply with CF conventions # np.arange(len(data_array))*1/self.fps
         # y needs to be flipped up down to match the order of rows followed by coordinate systems (bottom to top)
-        y = np.flipud(np.arange(data_array[0].shape[0]))
-        x = np.arange(data_array[0].shape[1])
+        # y = np.flipud(np.arange(data_array[0].shape[0]))
+        # x = np.arange(data_array[0].shape[1])
+        y = np.flipud(np.arange(sample.shape[0]))
+        x = np.arange(sample.shape[1])
         # perspective column and row coordinate grids
         xp, yp = np.meshgrid(x, y)
         coords = {
@@ -365,11 +499,12 @@ Camera configuration: {:s}
             "h_a": json.dumps(self.h_a)
         }
         frames = xr.DataArray(
-            da.stack(data_array, axis=0),
+            da_stack,
             dims=dims,
             coords=coords,
-            attrs=attrs
-        )
+            attrs=attrs,
+        )[::self.freq]
+        frames = frames.chunk({"time": 1})  # set chunks over time dimension
         del coords["time"]
         if len(sample.shape) == 3:
             del coords["rgb"]
@@ -378,57 +513,42 @@ Camera configuration: {:s}
         frames.name = "frames"
         return frames
 
+    def set_mask_from_exterior(
+            self,
+            exterior
+    ):
+        """
+        Prepare a mask grid with 255 outside of the stabilization polygon and 0 inside
 
-    def set_mask_from_exterior(self, exterior):
+        Parameters
+        ----------
+        exterior : list of lists
+            coordinates defining the polygon for masking
+
+        Returns
+        -------
+        self.mask : np.ndarray
+            mask for stabilization region
+
+        """
         mask_coords = np.array([exterior], dtype=np.int32)
         mask = np.zeros((self.height, self.width), np.uint8)
-        self.mask = cv2.fillPoly(mask, [mask_coords], 255)
+        mask = cv2.fillPoly(mask, [mask_coords], 255)
+        mask[mask == 0] = 1
+        mask[mask == 255] = 0
+        mask[mask == 1] = 255
+        self.mask = mask
 
-    def _get_pos_feats(self, cap, split=2, recipe=const.CLASSIFY_STANDING_CAM):
-        # go through the entire set of frames to gather transformation matrices per frame (except for the first one)
-        # get the displacements of trackable features
-        positions, stats, errs = cv._get_displacements(
+    def get_ms(
+            self,
+            cap: cv2.VideoCapture,
+            split: Optional[int] = 2
+    ):
+        self.ms = cv._get_ms_gftt(
             cap,
             start_frame=self.start_frame,
             end_frame=self.end_frame,
             split=split,
-            mask=self.mask
+            mask=self.mask,
         )
-        # filter features that belong to actual camera movement
-        for recipe in recipe:
-            classes = cv._classify_displacements(positions, **recipe)
-            # select positions which are classified as water
-            positions = positions[:, classes, :]
-            stats = stats[:, classes]
-        self.feats_pos = positions
-        self.feats_stats = stats
-        self.feats_errs = errs
 
-    def _get_ms(self):
-        # retrieve the transformation matrices for stabilization
-        self.ms = cv._ms_from_displacements(self.feats_pos, self.feats_stats)
-
-
-    def plot_rigid_pts(self, ax=None, **kwargs):
-        """
-        Plots found rigid points (column, row) for stabilization and their path throughout the frames in time on an
-        axes object.
-
-
-        Parameters
-        ----------
-        ax : plt.axes object, optional
-            If None (default), use the current axes.
-        **kwargs : additional keyword arguments to `matplotlib.pyplot.scatter` wrapped Matplotlib function.
-
-
-        Returns
-        -------
-
-        """
-        assert self.feats_pos is not None, "No stabilization applied hence no rigid points available to plot"
-        if ax is None:
-            ax = plt.axes()
-        for t_p in np.swapaxes(self.feats_pos, 0, 1):
-            p = ax.scatter(t_p[:, 0], t_p[:, 1], c=np.linspace(0, 1, len(t_p)), **kwargs)
-        return p
\ No newline at end of file
