@@ -8,66 +8,40 @@ from shapely.geometry import Polygon, LineString
 from shapely.affinity import rotate
 from tqdm import tqdm
 from scipy.cluster.vq import vq, kmeans
+from scipy import optimize
 import operator
 
+# default distortion coefficients for no distortion
+DIST_COEFFS = [[0.], [0.], [0.], [0.], [0.]]
 
 # criteria for finding subpix corners
 criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
 
-def _classify_displacements(positions, method="kmeans", q_threshold=0.8, abs_threshold=None, op=operator.le):
+
+
+def _combine_m(m1, m2):
+    # extend to a 3x3 for matrix multiplication
+    _m1 = np.append(m1, np.array([[0., 0., 1.]]), axis=0)
+    _m2 = np.append(m2, np.array([[0., 0., 1.]]), axis=0)
+    m_combi = _m1.dot(_m2)[0:2]
+    return m_combi
+
+
+def _smooth(img, stride):
     """
-    Classifies set of displacements of points through time in two groups based on a difference measure. Can e.g. be used
-    to mask values that are not moving enough (in case one wishes to detect water) or values that move too much
-    (in case one wishes to detect rigid points for image stabilization).
+    Internal function to filter on too large differences from spatial mean
 
     Parameters
     ----------
-    positions : list of arrays
-        time sequences of x,y locations per points
-    method : str, optional
-        method to filter or split points into population. Currently implemented are "kmeans" (split using a simple
-        kmeans classification with two assumed groups), "std" (split using a standard deviation criterium), or "dist"
-        (using an absolute distance in pixels).
-    q_threshold: float (0-1), optional
-        tolerance percentile used for either the "dist" or "std" method (i.e. points that move more or less than the
-        distance, measured as provided quantile, based on all points in the population are in either one or the other
-        group). Default: 0.8 meaning that points that have a standard deviation smaller than (larger or equal than, if
-        other operator is selected) the 0.8 quantile of all standard deviations are returned.
-    abs_threshold: float (0-1), optional
-        tolerance absolute value used for either the "dist" or "std" method (i.e. points that move more or less than the
-        distance, measured as provided absolute threshold, based on all points in the population are in either one or
-        the other group). Overrules q_threshold when set.
-    op: operator, optional
-        type of operation to test point values (default: operator.ge)
+    img: image
+    stride: window edge size
 
     Returns
     -------
-    filter: np.ndarray 1D [length positions]
-        Boolean per position, defining if a point filters out as True or False given the set criterion.
-        time sequences of x,y locations per points, after filtering
+    img
     """
-    assert(method in ["kmeans", "std", "dist"]), f'Method must be "kmeans", "std" or "dist", but instead is {method}.'
-    if q_threshold is not None:
-        assert (0.99 > q_threshold > 0.01), \
-            f'q_threshold represents a quantile and must be between 0.01 and 0.99, {q_threshold} given '
-    if method in ["kmeans", "std"]:
-        test_variable = positions.std(axis=0).mean(axis=-1)
-    elif method == "dist":
-        distance_xy = positions[-1] - positions[0]
-        test_variable = (distance_xy[:, 0]**2 + distance_xy[:, 1]**2)**0.5
-    if method == "kmeans":
-        centroids, mean_value = kmeans(test_variable, 2)
-        clusters, distances = vq(test_variable, np.sort(centroids))
-        return clusters == 0
-    # if not kmeans, then follow the same route for "dist" or "std"
-    # derive tolerance quantile
-    if abs_threshold is None:
-        # tolerance from quantile in distribution
-        tolerance = np.quantile(test_variable, q_threshold)  # PARAMETER
-    else:
-        # tolerance as absolute value
-        tolerance = abs_threshold
-    return op(test_variable, tolerance)
+    blur = cv2.GaussianBlur(img.astype("float32"), (stride, stride), 0)
+    return blur
 
 
 def _convert_edge(img, stride_1, stride_2):
@@ -92,184 +66,119 @@ def _get_dist_coefs(k1):
     dist[0, 0] = k1
     return dist
 
-
-def _get_cam_mtx(height, width, c=2.0, f=1.0):
+def _get_cam_mtx(height, width, c=2.0, focal_length=None):
     """
     Get 3x3 camera matrix from lens parameters
 
     :param height: height of image from camera
     :param width: width of image from camera
     :param c: float, optical center (default: 2.)
-    :param f: float, focal length (default: 1.)
+    :param f: float, focal length (optional)
     :return: camera matrix, to be used by cv2.undistort
     """
     # define camera matrix
     mtx = np.eye(3, dtype=np.float32)
-    # mtx[0, 2] = width / c  # define center x
-    # mtx[1, 2] = height / c  # define center y
-    # mtx[0, 0] = f  # define focal length x
-    # mtx[1, 1] = f  # define focal length y
-
     mtx[0, 2] = width / c  # define center x
     mtx[1, 2] = height / c  # define center y
-    mtx[0, 0] = width  # define focal length x
-    mtx[1, 1] = width  # define focal length y
+    if focal_length is None:
+        mtx[0, 0] = width  # define focal length x
+        mtx[1, 1] = width  # define focal length y
+    else:
+        mtx[0, 0] = focal_length  # define focal length x
+        mtx[1, 1] = focal_length  # define focal length y
     return mtx
 
-def _get_displacements(cap, start_frame=0, end_frame=None, n_pts=None, split=2, mask=None):
-    """
-    compute displacements from trackable features found in start frame
 
-    Parameters
-    ----------
-    cap : cv2.Capture object
-        video object, opened with cv2
-    start_frame : int, optional
-        first frame to perform point displacement analysis (default : 0)
-    end_frame : int, optional
-        last frame to process (must be larger than start_frame). Default: None, meaning the last frame in the video will
-        be used).
-    n_pts : int, optional
-        Number of features to track. If not set, the square root of the amount of pixels of the frames will be used
-    split : int, optional
-        Number of regions to split the entire field of view in to find points equally spread, defaults to 2
-    mask : np.ndarray (2D), optional
-        if set, the areas that are one, are assumed to be region of interest and therefore masked out for finding points
-
-    Returns
-    -------
-    positions : np.ndarray [M x N x 2]
-        positions of the points from frame to frame with M the amount of frames, N the amount of points, and 2 the x, y
-        coordinates
-    status : np.ndarray [M x N]
-        status of tracking of points, normally 1 is expected, 0 means that tracking for point in given frame did not
-         yield results (see also
-         https://docs.opencv.org/4.x/dc/d6b/group__video__track.html#ga473e4b886d0bcc6b65831eb88ed93323)
-
-    """
+def _get_ms_gftt(cap, start_frame=0, end_frame=None, n_pts=None, split=2, mask=None, wdw=4):
     # set end_frame to last if not defined
     if end_frame is None:
         end_frame = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
 
+    m = np.eye(3)[0:2]
+    # m2 = np.eye(3)[0:2]
+    ms = []
+    # ms2 = []
+    m_key = copy.deepcopy(m)
     # get start frame and points
     cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
+    # determine the mount of frames that must be processed
+    n_frames = int(end_frame + 1) - int(start_frame)
 
     # Read first frame
-    _, prev = cap.read()
+    _, img_key = cap.read()
     # Convert frame to grayscale
-    prev_gray = cv2.cvtColor(prev, cv2.COLOR_BGR2GRAY)
-    if mask is not None:
-        prev_gray[mask > 0] = 0.
-
-    # prepare outputs
-    n_frames = int(end_frame + 1) - int(start_frame)
-    transforms = np.zeros((n_frames - 1, 3), np.float32)
-
+    img1 = cv2.cvtColor(img_key, cv2.COLOR_BGR2GRAY)
+    img_key = img1
 
     if n_pts is None:
         # use the square root of nr of pixels in a frame to decide on n_pts
-        n_pts = int(np.sqrt(len(prev_gray.flatten())))
+        n_pts = int(np.sqrt(len(img_key.flatten())))
 
-    # split image in smaller chunks if user wants
-    v = 0
-    h = 0
-    ver_split, hor_split = np.int16(np.ceil(np.array(prev_gray.shape) / split))
-    prev_pts = np.zeros((0, 1, 2), np.float32)
-    while v < prev_gray.shape[0]:
-        while h < prev_gray.shape[1]:
-            sub_img = prev_gray[v:v + ver_split, h:h + hor_split]
-            # get points over several quadrants
-            subimg_pts = cv2.goodFeaturesToTrack(
-                sub_img,
-                maxCorners=int(n_pts/split**2),
-                qualityLevel=0.3,
-                minDistance=10,
-                blockSize=1
-            )
-            # add offsets for quadrants
-            if subimg_pts is not None:
-                subimg_pts[:, :, 0] += h
-                subimg_pts[:, :, 1] += v
-                prev_pts = np.append(prev_pts, subimg_pts, axis=0)
-            h += hor_split
-        h = 0
-        v += ver_split
-    # # get points over several quadrants
-    # prev_pts = cv2.goodFeaturesToTrack(
-    #     prev_gray,
-    #     maxCorners=n_pts,
-    #     qualityLevel=0.1,
-    #     minDistance=10,
-    #     blockSize=3
-    # )
-    # add start point to matrix
+    # get features from first key frame
+    prev_pts = _gftt_split(img_key, split, n_pts, mask=mask)
 
-    # get another frame a little bit ahead in time
-    # get start frame and points
-    # cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame + 30)
-    #
-    # success, curr = cap.read()
-    # Convert to grayscale
-    # curr_gray = cv2.cvtColor(curr, cv2.COLOR_BGR2GRAY)
-
-    # # Calculate optical flow
-    # curr_pts, status, err = cv2.calcOpticalFlowPyrLK(prev_gray, curr_gray, prev_pts, None)
-    # # Calculate optical flow
-    # curr_pts, status, err = cv2.calcOpticalFlowPyrLK(prev_gray, curr_gray, prev_pts, None)
-
-    # # store curr_pts
-    # positions = np.append(positions, np.swapaxes(curr_pts, 0, 1), axis=0)
-    # stats = np.append(stats, np.swapaxes(status, 0, 1), axis=0)
-    # errs = np.append(errs, np.swapaxes(err, 0, 1), axis=0)
-
-    # import matplotlib.pyplot as plt
-    # import matplotlib
-    # matplotlib.use("Qt5Agg")
-    # plt.imshow(prev_gray, cmap="Greys_r")
-    # x = prev_pts[:, :, 0].flatten()
-    # y = prev_pts[:, :, 1].flatten()
-    # intensity = prev_gray[np.int16(y), np.int16(x)]
-    # # x_curr = prev_pts[intensity<100, :, 0].flatten()
-    # # y_curr = prev_pts[intensity<100, :, 1].flatten()
-    # # x_curr = curr_pts[status.flatten()==1, :, 0].flatten()
-    # # y_curr = curr_pts[status.flatten()==1, :, 1].flatten()
-    # # prev_pts = prev_pts[intensity < 100]
-    #
-    # plt.plot(x, y, ".")
-    # # # plt.plot(x[status.flatten()==0], y[status.flatten()==0], "ro")
-    # # plt.plot(x_curr, y_curr, "r.")
-    # plt.show()
-    positions = np.swapaxes(prev_pts, 0, 1)
-    # update n_pts to the amount truly found
-    n_pts = positions.shape[1]
-    # prepare storage for points
-    # positions = np.zeros((0, n_pts, 2), np.float32)
-    stats = np.ones((1, n_pts))
-    errs = np.ones((1, n_pts))
-    # loop through start to end frame
-    pbar = tqdm(range(n_frames - 1), position=0, leave=True)
+    pbar = tqdm(range(n_frames), position=0, leave=True)
     for i in pbar:
-        # Read next frame
-        pbar.set_description(f"Stabilizing frames")
-        success, curr = cap.read()
-        if not success:
-            raise IOError(f"Could not read frame {start_frame + i} from video")
+        ms.append(m)
+        #     ms2.append(m2)
+        _, img2 = cap.read()
+        img2 = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)
+        curr_pts, status, err = cv2.calcOpticalFlowPyrLK(img_key, img2, prev_pts, None)
+        #     curr_pts = curr_pts[status == 1]
+        #     prev_pts = prev_pts[status == 1]
+        m_part = cv2.estimateAffine2D(curr_pts, prev_pts)[0]
+        m = _combine_m(m_key, m_part)
+        if i % 30 == 0:
+            img_key = img1
+            prev_pts = _gftt_split(img_key, split, n_pts, mask=mask)
+            m_key = copy.deepcopy(m)
+        img1 = img2
 
-        # Convert to grayscale
-        curr_gray = cv2.cvtColor(curr, cv2.COLOR_BGR2GRAY)
-        if mask is not None:
-            curr_gray[mask > 0] = 0.
-        # Calculate optical flow
-        curr_pts, status, err = cv2.calcOpticalFlowPyrLK(prev_gray, curr_gray, prev_pts, None)
+    # smooth the affines over time
+    ma = np.array(ms)
+    for m in range(ma.shape[1]):
+        for n in range(ma.shape[2]):
+            ma[wdw:-wdw, m, n] = np.convolve(ma[:, m, n], np.ones(wdw * 2 + 1) / (wdw * 2 + 1), mode="valid")
+    ms_smooth = list(ma)
+    return ms_smooth
 
-        # store curr_pts
-        positions = np.append(positions, np.swapaxes(curr_pts, 0, 1), axis=0)
-        stats = np.append(stats, np.swapaxes(status, 0, 1), axis=0)
-        errs = np.append(errs, np.swapaxes(err, 0, 1), axis=0)
-        # prepare next frame
-        prev_gray = curr_gray
-        prev_pts = curr_pts
-    return positions, stats, errs
+
+def _get_gcps_2_4(src, dst, img_width, img_height):
+    """
+
+    Parameters
+    ----------
+    src : list or array-like
+        source control points (list of lists)
+    dst : list or array-like
+        destination control points (list of lists)
+    img_width : width of original image frame
+    img_height : height of original image frame
+
+    Returns
+    -------
+    src : list or array-like
+        source control points (list of lists) converted to corner points
+    dst : list or array-like
+        destination control points (list of lists) converted to corner points
+    """
+    # first reverse the coordinate order of the y-axis
+    _src = [[x, img_height - y] for x, y in src]
+    # estimate affine transform (only rotation and translation)
+    M = cv2.estimateAffinePartial2D(np.array(_src), np.array(dst))
+    # complete M with a line indicating no perspective
+    M = np.array(M[0].tolist() + [[0, 0, 1]])
+    # establish corner coordinates
+    corners = [
+        [0, 0],
+        [img_width, 0],
+        [img_width, img_height],
+        [0, img_height]
+    ]
+    dst = cv2.perspectiveTransform(np.float32([corners]), M)[0].tolist()
+    # now get the corners back transformed to the real image coordinates
+    src = [[x, img_height - y] for x, y in corners]
+    return src, dst
 
 
 def _get_shape(bbox, resolution=0.01, round=1):
@@ -278,14 +187,14 @@ def _get_shape(bbox, resolution=0.01, round=1):
 
     :param bbox: shapely Polygon, bounding box
     :param resolution: resolution of target raster
-    :param round: number of pixels to round intended shape to
+    :param round: number of pixels to round intended sh ape to
     :return: numbers of rows and columns for target raster
     """
     coords = bbox.exterior.coords
     box_length = LineString(coords[0:2]).length
     box_width = LineString(coords[1:3]).length
-    cols = int(np.ceil((box_length / resolution) / round)) * round
-    rows = int(np.ceil((box_width / resolution) / round)) * round
+    cols = int(np.round((box_length / resolution) / round)) * round
+    rows = int(np.round((box_width / resolution) / round)) * round
     return cols, rows
 
 
@@ -325,138 +234,78 @@ def _get_transform(bbox, resolution=0.01):
     )
 
 
-def _get_gcps_a(lensPosition, h_a, coords, z_0=0.0, h_ref=0.0):
-    """Get the actual x, y locations of ground control points at the actual water level
+
+def _gftt_split(img, split, n_pts, mask=None):
+    # split image in smaller chunks if user wants
+    v = 0
+    h = 0
+    ver_split, hor_split = np.int16(np.ceil(np.array(img.shape) / split))
+    pts = np.zeros((0, 1, 2), np.float32)
+    while v < img.shape[0]:
+        while h < img.shape[1]:
+            sub_img = img[v:v + ver_split, h:h + hor_split]
+            # get points over several quadrants
+            subimg_pts = cv2.goodFeaturesToTrack(
+                sub_img,
+                mask=mask[v:v + ver_split, h:h + hor_split] if mask is not None else None,
+                maxCorners=int(n_pts/split**2),
+                qualityLevel=0.3,
+                minDistance=10,
+                blockSize=1
+            )
+            # add offsets for quadrants
+            if subimg_pts is not None:
+                subimg_pts[:, :, 0] += h
+                subimg_pts[:, :, 1] += v
+                pts = np.append(pts, subimg_pts, axis=0)
+            h += hor_split
+        h = 0
+        v += ver_split
+    return pts
+
+def solvepnp(dst, src, camera_matrix, dist_coeffs):
+    """
+    Short version with preprocessing for cv2.SolvePnP
 
     Parameters
     ----------
-    lensPosition : list of floats
-        x, y, z location of cam in local crs [m]
-    h_a : float
-        actual water level in local level measuring system [m]
-    coords : list of lists
-        gcp coordinates  [x, y] in original water level
-    z_0 : float, optional
-        reference zero plain level, i.e. the crs amount of meters of the zero level of staff gauge (default: 0.0)
-    h_ref : float, optional
-        reference water level during taking of gcp coords with ref to staff gauge zero level (default: 0.0)
+    src : list of lists
+        [x, y] with source coordinates, typically cols and rows in image
+    dst : list of lists
+        [x, y] (in case of 4 planar points) or [x, y, z] (in case of 6+ 3D points) with target coordinates after
+        reprojection, can e.g. be in crs [m]
+    camera_matrix : np.ndarray (3x3)
+        Camera intrinsic matrix
+    dist_coeffs : p.ndarray, optional
+        1xN array with distortion coefficients (N = 4, 5 or 8)
 
     Returns
     -------
-    coords : list
-        rows/cols for use in getPerspectivetransform
-
+    succes : int
+        0, 1 for succes or no succes
+    rvec : np.array
+        rotation vector
+    tvec : np.array
+        translation vector
     """
-    # get modified gcps based on camera location and elevation values
-    cam_x, cam_y, cam_z = lensPosition
-    x, y = zip(*coords)
-    # compute the z during gcp coordinates
-    z_ref = h_ref + z_0
-    # compute z during current frame
-    z_a = z_0 + h_a
-    # compute the water table to camera height difference during field referencing
-    cam_height_ref = cam_z - z_ref
-    # compute the actual water table to camera height difference
-    cam_height_a = cam_z - z_a
-    rel_diff = cam_height_a / cam_height_ref
-    # apply the diff on all coordinate, both x, and y directions
-    _dest_x = list(cam_x + (np.array(x) - cam_x) * rel_diff)
-    _dest_y = list(cam_y + (np.array(y) - cam_y) * rel_diff)
-    dest_out = list(zip(_dest_x, _dest_y))
-    return dest_out
+    # set points to float32
+    _src = np.float32(src)
+    _dst = np.float32(dst)
+
+    if len(_dst) == 4:
+        flags = cv2.SOLVEPNP_P3P
+        # if 4 x, y points are provided, add a column with zeros
+        # _dst = np.c_[_dst, np.zeros(4)]
+    else:
+        flags = cv2.SOLVEPNP_ITERATIVE
+
+    camera_matrix = np.float32(camera_matrix)
+    dist_coeffs = np.float32(dist_coeffs)
+    # define transformation matrix based on GCPs
+    return cv2.solvePnP(_dst, _src, camera_matrix, dist_coeffs, flags=flags)
 
 
-def m_from_displacement(p1, p2, status):
-    """
-    Calculate transform from pair of point locations, derived from Lukas Kanade optical flow.
-    The accompanying status array (see
-    https://docs.opencv.org/3.4/dc/d6b/group__video__track.html#ga473e4b886d0bcc6b65831eb88ed93323 is used to only
-    select points that were found in the optical flow algorithm.
-
-    Parameters
-    ----------
-    p1 : np.ndarray [n x 2]
-        point locations
-    p2 : np.ndarray [n x 2]
-        point locations (same as p1, but possibly displaced or rotated)
-
-    Returns
-    -------
-    m : affine matrix derived for 2D affine transform. Can be used with cv2.warpAffine
-
-    """
-    # remove any points that have a status zero according to optical flow
-    p1 = p1[status == 1]
-    p2 = p2[status == 1]
-    # add dim in the middle to match cv2 required array shape
-    prev_pts = np.float64(np.expand_dims(p1, 1))
-    curr_pts = np.float64(np.expand_dims(p2, 1))
-    return cv2.estimateAffinePartial2D(prev_pts, curr_pts)[0]
-
-
-def _ms_from_displacements(pts, stats, key_point=0, smooth=True):
-    """
-    Computes all transform matrices from list of point locations, found in frames, that are possibly moving.
-    The function returns transformation matrices that transform the position of all frames to one single frame (default
-    is the first frame)
-
-    Parameters
-    ----------
-    pts : np.ndarray [t x n x 2]
-        Location of traceable rigid body corner points (e.g. detected with good features to track and traced with
-        Lukas Kanade optical flow) through time
-        with t time steps, for n number of points, row column coordinates
-    stats : np.ndarray [t x n]
-        status of resolving optical flow (zero: not resolved, one: resolved, see
-        https://docs.opencv.org/3.4/dc/d6b/group__video__track.html#ga473e4b886d0bcc6b65831eb88ed93323
-    key_point : int, optional
-
-    Returns
-    -------
-    ms : list
-        Contains affine transform matrix for each set of points. This list can be used to affine transform each
-        frame to match as closely as possible the frame chosen as central frame.
-
-    """
-    assert key_point >= 0 and key_point < len(pts), f"Key point {int(key_point)} must be within range of point locations (0 - {len(pts) - 1}."
-    # TODO: approach to remove points before going to transformation is not working properly yet.
-    # remove points that at some point disappear, by finding which points have status that sometimes is zero
-    # idx = stats.min(axis=0) == 0
-    # pts = pts[:, idx, :]
-    # stats = stats[:, idx]
-
-    ms = [m_from_displacement(p2, pts[int(key_point)], status) for p2, status in zip(pts, stats)]
-    if smooth:
-        ms = _ms_smooth(ms)
-    return ms
-
-def _ms_smooth(ms, q=98):
-    def fill_1d(yp, xp):
-        idx = np.isfinite(yp)
-        yp_sel = yp[idx]
-        xp_sel = xp[idx]
-        return np.interp(xp, xp_sel, yp_sel)
-
-    ms = np.array(ms).reshape(len(ms), 6)
-
-    dms = np.diff(ms, axis=0)
-    tol = np.percentile(np.abs(dms), q, axis=0)
-
-    # remove values beyond the quyantile
-    dms[np.abs(dms) > tol] = np.nan
-
-    # also remove when one of the signals is nan
-    dms[np.isnan(dms.sum(axis=-1))] = np.nan
-
-    # fill the removed values with linear interpolated values
-    dms_fill = np.apply_along_axis(fill_1d, 0, dms, np.arange(len(dms)))
-    dms_fill = np.append(np.expand_dims(ms[0], 0), dms_fill, axis=0)
-
-    ms_out = list(np.cumsum(dms_fill, axis=0).reshape(len(ms), 2, 3))
-    return ms_out
-
-
-def _transform(img, m):
+def transform(img, m):
     """
     Affine transforms an image using a specified affine transform matrix. Typically the transformation is derived
     for image stabilization purposes.
@@ -585,6 +434,80 @@ def calibrate_camera(
     return camera_matrix, dist_coeffs
 
 
+def _Rt_to_M(rvec, tvec, camera_matrix, z=0., reverse=False):
+    R = cv2.Rodrigues(rvec)[0]
+    # assume height of projection plane
+    R[:, 2] = R[:, 2] * z
+    # add translation vector
+    R[:, 2] = R[:, 2] + tvec.flatten()
+    # compute homography
+    if reverse:
+        # From perspective to objective
+        M = np.dot(camera_matrix, R)
+    else:
+        # from objective to perspective
+        M = np.linalg.inv(np.dot(camera_matrix, R))
+    # normalize homography before returning
+    return M / M[-1, -1]
+
+def unproject_points(src, z, rvec, tvec, camera_matrix, dist_coeffs):
+    """
+    Reverse projects points from the image to the 3D world. As points on the objective are a ray line in
+    the real world, a x, y, z coordinate can only be reconstructed if the points have one known coordinate (z).
+
+    Parameters
+    ----------
+    src : list (of lists)
+        pixel coordinates
+    z : float or list of floats
+        z-level belonging to src points (if list, then the length should be equal to the number of src points)
+    rvec : array-like
+        rotation vector
+    tvec : array-like
+        translation vector
+    camera_matrix : array-like
+        camera matrix
+    dist_coeffs : array_like
+        distortion coefficients
+
+    Returns
+    -------
+
+    """
+    src = np.float32(np.atleast_1d(src))
+    # first undistort points
+    src = np.float32(
+        np.array(
+            undistort_points(
+                src,
+                camera_matrix,
+                dist_coeffs
+            )
+        )
+    )
+    if isinstance(z, (list, np.ndarray)):
+        #         zs = np.atleast_1d(zs)
+        z = np.float64(z)
+        dst = []
+        assert (len(z) == len(
+            src)), f"Amount of src points {len(src)} is not equal to amount of vertical levels z {len(z)}"
+        for pt, _z in zip(src, z):
+            M = _Rt_to_M(rvec, tvec, camera_matrix, z=_z, reverse=False)
+            x, y = list(cv2.perspectiveTransform(pt[None, None, ...], M)[0][0])
+            dst.append([x, y, _z])
+    else:
+        # z is only one value, so there is no change in M. We can do this faster in one go
+        M = _Rt_to_M(rvec, tvec, camera_matrix, z=z, reverse=False)
+        dst = cv2.perspectiveTransform(src[None, ...], M)[0]
+        dst = np.hstack(
+            (
+                dst,
+                np.ones((len(dst), 1)) * z,
+            )
+        )
+    return dst
+
+
 def get_M_2D(src, dst, reverse=False):
     """
     Retrieve homography matrix for between (4) src and (4) dst points with only x, y coordinates (no z)
@@ -613,6 +536,7 @@ def get_M_2D(src, dst, reverse=False):
         M = cv2.getPerspectiveTransform(_src, _dst)
     return M
 
+
 def get_M_3D(src, dst, camera_matrix, dist_coeffs=np.zeros((1, 4)), z=0., reverse=False):
     """
     Retrieve homography matrix for between (6+) 2D src and (6+) 3D dst (x, y, z) points
@@ -645,31 +569,125 @@ def get_M_3D(src, dst, camera_matrix, dist_coeffs=np.zeros((1, 4)), z=0., revers
     https://www.openearth.nl/flamingo/_modules/flamingo/rectification/rectification.html
 
     """
-    # set points to float32
-    _src = np.float32(src)
-    _dst = np.float32(dst)
-    # import pdb;pdb.set_trace()
-    camera_matrix = np.float32(camera_matrix)
-    dist_coeffs = np.float32(dist_coeffs)
-    # define transformation matrix based on GCPs
-    success, rvec, tvec = cv2.solvePnP(_dst, _src, camera_matrix, dist_coeffs)
-    # convert rotation vector to rotation matrix
-    R = cv2.Rodrigues(rvec)[0]
-    # assume height of projection plane
-    R[:, 2] = R[:, 2] * z
-    # add translation vector
-    R[:, 2] = R[:, 2] + tvec.flatten()
-    # compute homography
-    if reverse:
-        # From perspective to objective
-        M = np.dot(camera_matrix, R)
-    else:
-        # from objective to perspective
-        M = np.linalg.inv(np.dot(camera_matrix, R))
-    # normalize homography before returning
-    return M / M[-1, -1]
+    success, rvec, tvec = solvepnp(dst, src, camera_matrix, dist_coeffs)
+    return _Rt_to_M(rvec, tvec, camera_matrix, z=z, reverse=reverse)
 
-def get_time_frames(cap, start_frame, end_frame):
+def optimize_intrinsic(src, dst, height, width, c=2., lens_position=None):
+    def error_intrinsic(x, src, dst, height, width, c=2., lens_position=None, dist_coeffs=DIST_COEFFS):
+        """
+        estimate errors in known points using provided control on camera matrix.
+        returns error in gcps and camera position (if provided)
+        """
+        f = x[0]*width  # only one parameter to optimize for now, can easily be extended!
+        dist_coeffs[0][0] = float(x[1])
+        dist_coeffs[1][0] = float(x[2])
+        # dist_coeffs[4][0] = float(x[3])
+        # dist_coeffs[3][0] = float(x[4])
+        coord_mean = np.array(dst).mean(axis=0)
+        # _src = np.float32(src)
+        _dst = np.float32(np.array(dst) - coord_mean)
+        zs = np.zeros(4) if len(_dst[0]) == 2 else np.array(_dst)[:, -1]
+        if lens_position is not None:
+            _lens_pos = np.array(lens_position) - coord_mean
+
+        camera_matrix = _get_cam_mtx(height, width, c=c, focal_length=f)
+        success, rvec, tvec = solvepnp(_dst, src, camera_matrix, dist_coeffs)
+        if success:
+            # estimate destination locations from pose
+            dst_est = unproject_points(src, zs, rvec, tvec, camera_matrix, dist_coeffs)
+            # src_est = np.array([list(point[0]) for point in src_est])
+            dist_xy = np.array(_dst)[:, 0:2] - np.array(dst_est)[:, 0:2]
+            dist = (dist_xy ** 2).sum(axis=1) ** 0.5
+            gcp_err = dist.mean()
+            # print(f"Error: {gcp_err}")
+            # print(f"Parameters: {x}")
+            if lens_position is not None:
+                rmat = cv2.Rodrigues(rvec)[0]
+                lens_pos2 = np.array(-rmat).T @ tvec
+                cam_err = ((_lens_pos - lens_pos2.flatten()) ** 2).sum() ** 0.5
+            else:
+                cam_err = None
+                # TODO: for now camera errors are weighted with 10% needs further investigation
+            err = float(0.1*cam_err + gcp_err) if cam_err is not None else gcp_err
+        else:
+            err = 100
+        return err  # assuming gcp pixel distance is about 5 cm
+
+    if len(dst) == 4:
+        bnds_k1 = (-0.0, 0.0)
+        bnds_k2 = (-0.0, 0.0)
+    else:
+        # bnds_k1 = (-0.2501, -0.25)
+        bnds_k1 = (-0.9, 0.9)
+        bnds_k2 = (-0.5, 0.5)
+    opt = optimize.differential_evolution(
+        error_intrinsic,
+        # bounds=[(float(0.25), float(2)), bnds_k1],#, (-0.5, 0.5)],
+        bounds=[(float(0.25), float(2)), bnds_k1, bnds_k2],
+        # bounds=[(1710./width, 1714./width), bnds_k1, bnds_k2],
+        args=(src, dst, height, width, c, lens_position, DIST_COEFFS),
+        atol=0.001 # one mm
+    )
+    camera_matrix = _get_cam_mtx(height, width, focal_length=opt.x[0]*width)
+    dist_coeffs = DIST_COEFFS
+    dist_coeffs[0][0] = opt.x[1]
+    dist_coeffs[1][0] = opt.x[2]
+    # dist_coeffs[4][0] = opt.x[3]
+    # dist_coeffs[3][0] = opt.x[4]
+    print(f"CAMERA MATRIX: {camera_matrix}")
+    print(f"DIST COEFFS: {dist_coeffs}")
+    return camera_matrix, dist_coeffs, opt.fun
+
+
+def get_frame(
+        cap,
+        rotation=None,
+        ms=None,
+        method="grayscale"
+):
+    try:
+        ret, img = cap.read()
+        if rotation is not None:
+           img = cv2.rotate(img, rotation)
+    except:
+        raise IOError(f"Cannot read")
+    if ret:
+        if ms is not None:
+            img = transform(img, ms)
+        # apply lens distortion correction
+        if method == "grayscale":
+            # apply gray scaling, contrast- and gamma correction
+            # img = _corr_color(img, alpha=None, beta=None, gamma=0.4)
+            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # mean(axis=2)
+        elif method == "rgb":
+            # turn bgr to rgb for plotting purposes
+            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
+        elif method == "hsv":
+            img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
+    return ret, img
+
+def get_frames(cap, start_frame, end_frame):
+    """
+    Get a set of frames from start_frame to end frame from a cap object
+
+    Parameters
+    ----------
+    cap : cv.VideoCapture
+        Opened VideoCapture object
+    start_frame : int
+        first frame to consider for reading
+    end_frame : int
+        last frame to consider for reading
+
+    Returns
+    -------
+    frames : list
+        list with 2D or 3D (in case of RGB) numpy arrays with image intensity values
+    """
+    cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
+
+
+def get_time_frames(cap, start_frame, end_frame, lazy=True, fps=None, **kwargs):
     """
     Gets a list of valid time stamps and frame numbers for the provided video capture object, starting from start_frame
     ending at end_frame
@@ -682,6 +700,8 @@ def get_time_frames(cap, start_frame, end_frame):
             first frame to consider for reading
         end_frame : int
             last frame to consider for reading
+        fps : float
+            hard enforced frames per second number (used when metadata of video is incorrect)
 
         Returns
         -------
@@ -692,17 +712,28 @@ def get_time_frames(cap, start_frame, end_frame):
 
     """
     cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
-    ret, img = cap.read()  # read first frame
+    ret, img = get_frame(cap, **kwargs)
     n = start_frame
     time = []
     frame_number = []
+    if lazy:
+        frames = None
+    else:
+        # already collect the frames
+        frames = []
     while ret:
         if n > end_frame:
             break
+        if not(lazy):
+            frames.append(img)
         t1 = cap.get(cv2.CAP_PROP_POS_MSEC)
-        ret, img = cap.read()  # read frame 1 + ...
+        if fps is not None:
+            time.append(n*1000./fps)
+        else:
+            time.append(t1)
+        # ret, img = cap.read()  # read frame 1 + ...
+        ret, img = get_frame(cap, **kwargs)    # read frame 1 + ...
         frame_number.append(n)
-        time.append(t1)
         if ret == False:
             break
         # cv2.imwrite("test_{:04d}.jpg".format(n), img)
@@ -712,7 +743,8 @@ def get_time_frames(cap, start_frame, end_frame):
             break
 
         n += 1
-    return time, frame_number
+    # time[0] = 0
+    return time, frame_number, frames
 
 
 def transform_to_bbox(coords, bbox, resolution):
@@ -721,7 +753,7 @@ def transform_to_bbox(coords, bbox, resolution):
     Parameters
     ----------
     coords : list of lists
-        [x, y] with coordinates
+        [x, y, z] with coordinates
     bbox : shapely Polygon
         Bounding box. The coordinate order is very important and has to be upstream-left, downstream-left,
         downstream-right, upstream-right, upstream-left
@@ -735,10 +767,14 @@ def transform_to_bbox(coords, bbox, resolution):
 
     """
     # first assemble x and y coordinates
-    xs, ys = zip(*coords)
     transform = _get_transform(bbox, resolution)
+    if len(coords[0]) == 3:
+        xs, ys, zs = zip(*coords)
+    else:
+        xs, ys = zip(*coords)
     rows, cols = rasterio.transform.rowcol(transform, xs, ys, op=float)
-    return list(zip(cols, rows))
+    return list(zip(cols, rows)) if len(coords[0]) == 2 else list(zip(cols, rows, zs))
+
 
 
 def get_ortho(img, M, shape, flags=cv2.INTER_AREA):
@@ -764,32 +800,27 @@ def get_ortho(img, M, shape, flags=cv2.INTER_AREA):
     return cv2.warpPerspective(img, M, shape, flags=flags)
 
 
-def get_aoi(M, src_corners, resolution=None):
+def get_aoi(dst_corners, resolution=None):
     """Get rectangular AOI from 4 user defined points within frames.
 
     Parameters
     ----------
-    src : list of tuples
-        (col, row) pairs of ground control points
-    dst : list of tuples
-        projected (x, y) coordinates of ground control points
+    M : np.ndarray
+        Homography matrix
     src_corners : dict with 4 (x,y) tuples
         names "up_left", "down_left", "up_right", "down_right", source corners
+    resolution : float
+        resolution of intended reprojection, used to round the bbox to a whole number of intended pixels
 
     Returns
     -------
-    aoi : shapely.geometry.Polygon
-        bounding box of aoi (rotated)
+    bbox : shapely.geometry.Polygon
+        bounding box of aoi (with rotated affine)
     """
-    # retrieve the M transformation matrix for the conditions during GCP. These are used to define the AOI so that
-    # dst AOI remains the same for any movie
     # prepare a simple temporary np.array of the src_corners
-    _src_corners = np.array(src_corners)
-    assert(_src_corners.shape==(4, 2)), f"a list of lists of 4 coordinates must be given, resulting in (4, 2) shape. " \
-                                        f"Current shape is {src_corners.shape} "
-    # reproject corner points to the actual space in coordinates
-    _dst_corners = cv2.perspectiveTransform(np.float32([_src_corners]), M)[0]
-    polygon = Polygon(_dst_corners)
+    # assert(_src_corners.shape==(4, 2)), f"a list of lists of 4 coordinates must be given, resulting in (4, 2) shape. " \
+    #                                     f"Current shape is {src_corners.shape} "
+    polygon = Polygon(dst_corners)
     coords = np.array(polygon.exterior.coords)
     # estimate the angle of the bounding box
     # retrieve average line across AOI
@@ -799,7 +830,7 @@ def get_aoi(M, src_corners, resolution=None):
     angle = np.arctan2(diff[1], diff[0])
     # rotate the polygon over this angle to get a proper bounding box
     polygon_rotate = rotate(
-        polygon, -angle, origin=tuple(_dst_corners[0]), use_radians=True
+        polygon, -angle, origin=tuple(dst_corners[0]), use_radians=True
     )
     xmin, ymin, xmax, ymax = polygon_rotate.bounds
     if resolution is not None:
@@ -811,7 +842,7 @@ def get_aoi(M, src_corners, resolution=None):
     bbox_coords = [(xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin), (xmin, ymax)]
     bbox = Polygon(bbox_coords)
     # now rotate back
-    bbox = rotate(bbox, angle, origin=tuple(_dst_corners[0]), use_radians=True)
+    bbox = rotate(bbox, angle, origin=tuple(dst_corners[0]), use_radians=True)
     return bbox
 
 
@@ -836,13 +867,6 @@ def undistort_img(img, camera_matrix, dist_coeffs):
         undistorted img
     """
 
-    # define imagery characteristics
-    # height, width, __ = img.shape
-    # dist = _get_dist_coefs(k1)
-    #
-    # # get camera matrix
-    # mtx = _get_cam_mtx(height, width, c=c, f=f)
-
     # correct image for lens distortion
     return cv2.undistort(img, np.array(camera_matrix), np.array(dist_coeffs))
 
