@@ -1,19 +1,19 @@
 import copy
+import click
 import functools
 import logging
 import os.path
-
-import click
-
 import pyorc
+import subprocess
 import xarray as xr
 import yaml
-from pyorc.cli import cli_utils
 
+from pyorc.cli import cli_utils
 from dask.diagnostics import ProgressBar
 from matplotlib.colors import Normalize
 from typing import Dict
 
+__all__ = ["velocity_flow", "velocity_flow_subprocess"]
 
 logger = logging.getLogger(__name__)
 
@@ -100,7 +100,7 @@ def run_func_hash_io(
                 if not (os.path.isfile(fn_recipe)):
                     run = True
                 else:
-                    recipe_part = {c: ref.recipe[c] for c in configs}
+                    recipe_part = {c: ref.recipe[c] for c in configs if c in ref.recipe}
                     with open(fn_recipe, "r") as f:
                         cfg_ancient = f.read()
                     cfg = yaml.dump(recipe_part, default_flow_style=False, sort_keys=False)
@@ -136,7 +136,7 @@ def run_func_hash_io(
                 processor_func(ref, *args, **kwargs)
                 # after run, store configuration file and hashes of in- and outputs
                 fn_recipe = os.path.join(path_out, f"{ref.prefix}{func_name}.yml")
-                recipe_part = {c: ref.recipe[c] for c in configs}
+                recipe_part = {c: ref.recipe[c] for c in configs if c in ref.recipe}
                 with open(fn_recipe, "w") as f:
                     yaml.dump(recipe_part, f, default_flow_style=False, sort_keys=False)
                 # after run, store input and output hashes
@@ -168,10 +168,12 @@ class VelocityFlowProcessor(object):
             self,
             recipe: Dict,
             videofile: str,
-            cameraconfig: str,
+            cameraconfig: Dict,
             prefix: str,
             output: str,
+            h_a: float=None,
             update: bool=False,
+            concurrency=True,
             fn_piv="piv.nc",
             fn_piv_mask="piv_mask.nc",
             fn_transect_template="transect_{:s}.nc",
@@ -186,19 +188,24 @@ class VelocityFlowProcessor(object):
             YAML recipe, parsed from CLI
         videofile : str
             path to video
-        cameraconfig : str
-            path to camera config file
+        cameraconfig : dict
+            camera config as dict (not yet loaded as CamerConfig object)
         prefix : str
             prefix of produced output files
         output : str
             path to output file
         update : bool, optional
             if set, only update components with changed inputs and configurations
+        concurrency : bool, optional
+            if set to False, then dask will only run synchronous preventing overuse of memory. This will be slower
 
         """
+        if h_a is not None:
+            recipe["video"]["h_a"] = h_a
         self.update = update  # set to True when checks are needed if data already exists or not
         self.recipe = recipe
         self.output = output
+        self.concurrency = concurrency
         self.prefix = prefix
         self.fn_piv = os.path.join(self.output, prefix + fn_piv)
         self.fn_piv_mask = os.path.join(self.output, prefix + fn_piv_mask) if "mask" in recipe else self.fn_piv
@@ -208,10 +215,8 @@ class VelocityFlowProcessor(object):
         self.read = True
         self.write = False
         self.fn_video = videofile
-        self.fn_cam_config = cameraconfig
+        self.cam_config = pyorc.CameraConfig(**cameraconfig)
         self.logger = logger
-        # self.set_status_fn(stat_fn)
-        # self.get_status()
         # TODO: perform checks, minimum steps required
         self.logger.info("pyorc velocimetry processor initialized")
 
@@ -256,30 +261,6 @@ class VelocityFlowProcessor(object):
         self._fn_video = fn_video
 
 
-    def set_status_fn(self, fn):
-        """
-        Prepare expected status file, containing filenames and hashes of existing files if these are already processed
-
-        """
-        self.status_fn = os.path.join(self.output, fn)
-
-
-    def get_status(self):
-        """
-        sets status of project
-        Returns
-        -------
-
-        """
-        if os.path.isfile(self.status_fn):
-            # read file and return dict
-            with open(self.status_fn, "r") as f:
-                body = f.read()
-            self.status = yaml.load(body, Loader=yaml.FullLoader)
-        else:
-            self.status = {}
-
-
     def process(self):
         """
         Single method to perform all processing in logical pre-defined order. Also checks for compulsory steps
@@ -289,7 +270,12 @@ class VelocityFlowProcessor(object):
         -------
 
         """
-
+        if not(self.concurrency):
+            import dask
+            # run only synchronous
+            dask.config.set(scheduler='synchronous')
+        # dask.config.set(pool=Pool(4))
+        # dask.config.set(scheduler='processes')
         self.video(**self.recipe["video"])
         self.frames(**self.recipe["frames"])
         self.velocimetry(**self.recipe["velocimetry"])
@@ -300,12 +286,20 @@ class VelocityFlowProcessor(object):
             self.velocimetry_mask_obj = self.velocimetry_obj
         if "transect" in self.recipe:
             self.transect(**self.recipe["transect"])
-        else:
-            # no masking so use non-masked velocimetry as masked
-            self.velocimetry_mask_obj = self.velocimetry_obj
+        # else:
+        #     # no masking so use non-masked velocimetry as masked
+        #     self.velocimetry_mask_obj = self.velocimetry_obj
         if "plot" in self.recipe:
             self.plot(**self.recipe["plot"])
-
+        # remove all potentially memory consumptive attributes
+        self.da_frames.close()
+        self.velocimetry_mask_obj.close()
+        self.velocimetry_obj.close()
+        delattr(self, "video_obj")
+        delattr(self, "velocimetry_obj")
+        delattr(self, "velocimetry_mask_obj")
+        delattr(self, "da_frames")
+        return None
         # TODO .get_transect and check if it contains data,
 
         #  perform any post processing such as plotting or possibly later other analyses
@@ -319,7 +313,7 @@ class VelocityFlowProcessor(object):
     def video(self, **kwargs):
         self.video_obj = pyorc.Video(
             self.fn_video,
-            camera_config=self.fn_cam_config,
+            camera_config=self.cam_config,
             **kwargs
         )
         # some checks ...
@@ -345,7 +339,6 @@ class VelocityFlowProcessor(object):
             kwargs_video = kwargs["to_video"]
             self.logger.info(f"Writing video of processed frames to {kwargs_video['fn']}")
             self.da_frames.frames.to_video(**kwargs_video)
-
         self.logger.info(f'Frames are preprocessed')
 
 
@@ -364,12 +357,14 @@ class VelocityFlowProcessor(object):
             delayed_obj = self.velocimetry_obj.to_netcdf(self.fn_piv, compute=False)
             with ProgressBar():
                 delayed_obj.compute()
+            del delayed_obj
             self.logger.info(f"Velocimetry written to {self.fn_piv}")
             # Load the velocimetry into memory to prevent re-writes in next steps
+            delattr(self, "velocimetry_obj")
             self.velocimetry_obj = xr.open_dataset(self.fn_piv)
 
 
-    @run_func_hash_io(attrs=["velocimetry_mask_obj"], check=True, inputs=["fn_piv"], outputs=["fn_piv_mask"])
+    @run_func_hash_io(attrs=["velocimetry_mask_obj"], check=True, inputs=["fn_piv"], configs=["video", "frames", "velocimetry", "mask"], outputs=["fn_piv_mask"])
     def mask(self, write=False, **kwargs):
         # TODO go through several masking groups
         self.velocimetry_mask_obj = copy.deepcopy(self.velocimetry_obj)
@@ -386,6 +381,7 @@ class VelocityFlowProcessor(object):
             delayed_obj = self.velocimetry_mask_obj.to_netcdf(self.fn_piv_mask, compute=False)
             with ProgressBar():
                 delayed_obj.compute()
+            del delayed_obj
 
 
     @run_func_hash_io(check=False, configs=["transect"], inputs=["fn_piv_mask"])
@@ -395,10 +391,17 @@ class VelocityFlowProcessor(object):
         _kwargs = copy.deepcopy(kwargs)
         for transect_name, transect_grp in _kwargs.items():
             self.logger.debug(f'Processing transect "{transect_name}"')
-            if not "shapefile" in transect_grp:
-                raise click.UsageError(f'Transect with name "{transect_name}" does not have a "shapefile". Please add "shapefile" in the recipe file')
-            # read shapefile
-            coords, crs = cli_utils.read_shape(transect_grp["shapefile"])
+            # check if there are coordinates provided
+
+            if not ("shapefile" in transect_grp or "geojson" in transect_grp):
+                raise click.UsageError(f'Transect with name "{transect_name}" does not have a "shapefile" or '
+                                       f'"geojson". Please add "shapefile" in the recipe file')
+            # read geojson or shapefile (as alternative
+            if "geojson" in transect_grp:
+                # read directly from geojson
+                coords, crs = cli_utils.read_shape(geojson=transect_grp["geojson"])
+            elif "shapefile" in transect_grp:
+                coords, crs = cli_utils.read_shape(fn=transect_grp["shapefile"])
             self.logger.debug(f"Coordinates read for transect {transect_name}")
             # check if coords have z coordinates
             if len(coords[0]) == 2:
@@ -443,7 +446,7 @@ class VelocityFlowProcessor(object):
                 self.logger.info(f'Transect "{transect_name}" written to {fn_transect}')
 
 
-    @run_func_hash_io(check=True, configs=["video", "frames", "velocimetry", "transect", "plot"], inputs=["fn_video", "fn_piv_mask"], outputs=[])
+    @run_func_hash_io(check=False, configs=["video", "frames", "velocimetry", "transect", "plot"], inputs=["fn_video", "fn_piv_mask"], outputs=[])
     def plot(self, **plot_recipes):
         _plot_recipes = copy.deepcopy(plot_recipes)
         for name, plot_params in _plot_recipes.items():
@@ -483,6 +486,7 @@ class VelocityFlowProcessor(object):
                 )
                 p = velocimetry_reduced.velocimetry.plot(ax=ax, mode=mode, **opts)
                 ax = p.axes
+                del velocimetry_reduced
             if "transect" in plot_params:
                 for transect_name, opts in plot_params["transect"].items():
                     opts = vmin_vmax_to_norm(opts)
@@ -495,6 +499,9 @@ class VelocityFlowProcessor(object):
                     # add to plot
                     p = ds_trans_q.transect.plot(ax=ax, mode=mode, **opts)
                     ax = p.axes
+                    # done with transect, remove from memory
+                    ds_trans.close()
+                    del ds_trans
             # if mode == "camera":
             #     ax.axis("equal")
             write_pars = plot_params["write_pars"] if "write_pars" in plot_params else {}
@@ -502,3 +509,87 @@ class VelocityFlowProcessor(object):
             ax.figure.savefig(fn_jpg, **write_pars)
             self.logger.info(f'Plot "{name}" written to {fn_jpg}')
 
+def velocity_flow(**kwargs):
+    # simply execute the entire process
+    processor = VelocityFlowProcessor(**kwargs)
+    # process video following the settings
+    processor.process()
+    # flush the processor itself once done
+    del processor
+
+
+def velocity_flow_subprocess(
+    videofile,
+    recipe,
+    cameraconfig,
+    output,
+    prefix=None,
+    h_a: float = None,
+    update: bool = False,
+    concurrency=True,
+    logger=logging
+):
+    """
+    Writes the requirements to temporary files and runs velocimetry from a separate CLI instance
+
+    Parameters
+    ----------
+    recipe : dict
+        YAML recipe, parsed from CLI
+    videofile : str
+        path to video
+    cameraconfig : dict
+        camera config as dict (not yet loaded as CamerConfig object)
+    prefix : str
+        prefix of produced output files
+    output : str
+        path to output file
+    update : bool, optional
+        if set, only update components with changed inputs and configurations
+    concurrency : bool, optional
+        if set to False, then dask will only run synchronous preventing overuse of memory. This will be slower
+
+
+    Returns
+    -------
+
+    """
+    # store recipe in file
+    logger.info(f"Launching separate pyorc instance for videofile {videofile}")
+    if not(os.path.isdir(output)):
+        os.makedirs(output)
+    fn_recipe = os.path.join(output, "recipe.yml")
+    fn_cam_config = os.path.join(output, "camera_config.json")
+    with open(fn_recipe, "w") as f:
+        yaml.dump(recipe, f, default_flow_style=False, sort_keys=False)
+    pyorc.CameraConfig(**cameraconfig).to_file(fn_cam_config)
+    cmd = [
+        "pyorc",
+        "velocimetry",
+        "-V",
+        videofile,
+        "-c",
+        fn_cam_config,
+        "-r",
+        fn_recipe
+    ]
+    # add components where needed
+    if h_a is not None:
+        cmd.append("-h")
+        cmd.append(str(h_a))
+    if concurrency == False:
+        cmd.append("--lowmem")
+    if update:
+        cmd.append("-u")
+    if prefix:
+        cmd.append("-p")
+        cmd.append(prefix)
+    cmd_suffix = [
+        "-u",
+        "-vvv",
+        output
+    ]
+    cmd = cmd + cmd_suffix
+    # call subprocess
+    result = subprocess.run(cmd)
+    return result
\ No newline at end of file
