@@ -6,8 +6,8 @@ def test_get_river_flow(piv_transect):
     # fill method is already tested in get_q, so choose default only
     piv_transect.transect.get_q()
     piv_transect.transect.get_river_flow()
-    # because we only have PIV for one time step, all quantiles will have the same values
-    assert(np.allclose(piv_transect.river_flow.values, [0.07360698, 0.0772654 , 0.08183843, 0.08641146, 0.09006988]))
+    # we allow for 0.001 m3/s deviation for differences in versions of libs
+    assert(np.allclose(piv_transect.river_flow.values, [0.12069583, 0.12513554, 0.13068518, 0.13623481, 0.14067452], atol=0.001))
 
 
 @pytest.mark.parametrize(
@@ -36,10 +36,9 @@ def test_get_q(piv_transect, fill_method):
     "method",
     [
         "quiver",
-        "scatter",
+        # "scatter",
     ]
 )
 def test_plot(piv_transect, mode, method):
     piv_transect.transect.get_q()
-    piv_transect.isel(quantile=2).transect.plot(method=method, mode=mode)
-    plt.close("all")
+    piv_transect.isel(quantile=2).transect.plot(method=method, mode=mode, add_text=True)
